import os
import numpy as np
import csv
from scipy.spatial import KDTree

try:
    from pathlib import Path
    Path().expanduser()
except (ImportError,AttributeError):
    from pathlib2 import Path

try:
    import tempfile
    tempfile.TemporaryDirectory

except (ImportError,AttributeError):
    from backports import tempfile









def compute_nearest(box, boxes, secondboxes, mindist = 20):
    
    xbox = box[4]
    ybox = box[5]
    
    
    pts = []
    if boxes is not None:
     for i in range(0, boxes.shape[0]):
        xboxes = boxes[i,4]
        yboxes = boxes[i,5]
        center = (xboxes, yboxes)
        pts.append(center)
    secondpts = []
    if secondboxes is not None:
     for i in range(0, secondboxes.shape[0]):
        xboxes = secondboxes[i,4]
        yboxes = secondboxes[i,5]
        center = (xboxes, yboxes)
        secondpts.append(center)    
    indexA = None
    indexB = None    
    if len(pts)> 0:   
      pts = np.array(pts)
      Target = KDTree(pts)
      if Target is not None:     
        idx = Target.query_ball_point([xbox, ybox], r = mindist)
        if idx is not None :
         #Return the list of nearest points
         Nearest = pts[idx]
         SortNearest = np.sort(Nearest[::-1])
         if len(SortNearest)>1: 
          NearestXY = SortNearest[0]
          if boxes is not None:   
           for i in range(0, boxes.shape[0]):
             if(boxes[i,4] == NearestXY[0] and boxes[i,5] == NearestXY[1] ):
                indexA = i
      if len(secondpts)> 0:
       secondpts = np.array(secondpts)
       secondTarget = KDTree(secondpts)
       #Get id array of all nearest points

       if secondTarget is not None: 
        secondidx = secondTarget.query_ball_point([xbox, ybox], r = mindist)
 
       if secondidx is not None: 
        secondNearest = secondpts[secondidx]
        #Sort the list of nearest points
     
        SortsecondNearest = np.sort(secondNearest[::-1])

        if len(SortsecondNearest>1): 
         secondNearestXY = SortsecondNearest[0] 
       
         if secondboxes is not None:      
          for i in range(0, secondboxes.shape[0]):
           if(secondboxes[i,4] == secondNearestXY[0] and secondboxes[i,5] == secondNearestXY[1]):
              indexB = i

 
       
    return indexA, indexB  





def drawimage(eventlist, basedirResults, fname, Label):
   if eventlist is not None: 
    
      
    
     LocationXlist = []
     LocationYlist = []
     Timelist = []
    
     for i in range(0, len(eventlist)):
                    location,time,Name = eventlist[i]
                    (x,y) = location
                    returntime = int(time) 
                    
                    
                    LocationXlist.append(location[0])
                    LocationYlist.append(location[1])
                    Timelist.append(returntime)
                  
     Event_Count = np.column_stack([LocationXlist, LocationYlist,Timelist]) 
     Event_data = []
     
     
     for line in Event_Count:
        Event_data.append(line)
        writer = csv.writer(open(basedirResults + "/" + str(Label) + "LocationEventCounts" + (os.path.splitext(os.path.basename(fname))[0])  +".csv", "a"))
        writer.writerows(Event_data)
        Event_data = []
    


                 

                        



def StaticEvents(Image,  timelist, TimeFrames, boxes, originalsize, threshold, timepoint ,basedirResults, axes, fname, AppendName):
    
    originalsizeX, originalsizeY = originalsize
    
    
    EventList, EventBoxes = NMSSpace(boxes, originalsizeX, originalsizeY, threshold)
    

    
    return [], EventList
    
def DynamicEvents(Image,  timelist, TimeFrames, boxes, originalsize, threshold, timepoint, basedirResults, axes, fname, AppendName):
    
    originalsizeX, originalsizeY = originalsize
    EventList, boxes = NMSSpace(boxes, originalsizeX, originalsizeY, threshold)
    if len(EventList) > 0:
     boxes = boxes.tolist()
    EventList = []
    if timepoint%(TimeFrames + 2) == 0 and timepoint!=0:
        
      EventList, EventBoxes = NMSSpace(boxes, originalsizeX, originalsizeY, threshold)
      
      boxes = []
      
    
    return boxes, EventList  
    
def NMSSpace(boxes, originalsizeX, originalsizeY,threshold):
    
   if boxes is not None: 
    if len(boxes) == 0:
      return [], []
    else:
     boxes = np.array(boxes, dtype = float)
     assert boxes.shape[0] > 0
     if boxes.dtype.kind != "f":
        boxes = boxes.astype(np.float32)
     
     score = -boxes[:,6]
     idxs = np.argsort(score)

     pick = []
     while len(idxs) > 0:
        i = idxs[0]
        pick.append(i)
 
        
        distance = compute_dist(boxes[i], boxes[idxs[1:]])
      
        remove_idxs = np.where(distance < threshold)[0] + 1
        idxs = np.delete(idxs, remove_idxs)
        idxs = np.delete(idxs, 0)
        
     centerlist = []    

     for i in range(len(boxes[pick,0])):
 
       center = ( ((boxes[pick,4][i])) , ((boxes[pick,5][i])) )
       centerlist.append([center, boxes[pick,7][i],boxes[pick,8][i]] )
     if(boxes[pick,8][0] == 1):
       Name = 'Apoptosis'
     if(boxes[pick,8][0] == 2):
       Name = 'Division'
     if(boxes[pick,8][0] == 3):
       Name = 'MacroKitty'
     if(boxes[pick,8][0] == 4):
       Name = 'Non-Mature'
     if(boxes[pick,8][0] == 5):
       Name = 'Mature'
     print('Number of events:', len(centerlist),Name)

    
     return centerlist, boxes[pick]





def compute_dist(box, boxes):
    # Calculate intersection areas
    
    Xtarget = boxes[:, 4]
    Ytarget = boxes[:, 5]
    Ttarget = boxes[:, 7]
    
    Xsource = box[4]
    Ysource = box[5]
    Tsource = box[7]
    
    # If seperated in time the distance is made lower to avoid multi counting of events
    
        
    distance = (Xtarget - Xsource) * (Xtarget - Xsource) + (Ytarget - Ysource) * (Ytarget - Ysource) - (Ttarget - Tsource) * (Ttarget - Tsource)
    
    distance[ distance < 0] = 0
    
    return np.sqrt(distance)









#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  4 14:50:47 2021

@author: vkapoor
"""
from tifffile import imread, imwrite
import csv
import napari
import glob
import os
import cv2
import random
import sys
import numpy as np
import json
from scipy import spatial
from pathlib import Path
from scipy import spatial
import itertools
from napari.qt.threading import thread_worker
import matplotlib.pyplot  as plt
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QComboBox, QPushButton, QSlider
import h5py
from skimage import measure
from scipy.ndimage import gaussian_filter
import pandas as pd
import imageio
from oneat.NEATUtils.napari_animation._qt import AnimationWidget
from dask.array.image import imread as daskread
Boxname = 'ImageIDBox'
EventBoxname = 'EventIDBox'


class NEATViz(object):

        def __init__(self, imagedir, heatmapimagedir, savedir, categories_json, event_threshold, heatname = '_Heat', fileextension = '*tif', blur_radius = 5, start_project_mid = 0, end_project_mid = 0 ):
            
            
               self.imagedir = imagedir
               self.heatmapimagedir = heatmapimagedir
               self.savedir = savedir
               self.heatname = heatname
               self.event_threshold = event_threshold
               self.categories_json = categories_json
               self.start_project_mid = start_project_mid
               self.end_project_mid = end_project_mid
               self.fileextension = fileextension
               self.blur_radius = blur_radius
               Path(self.savedir).mkdir(exist_ok=True)
               self.viewer = napari.Viewer()
               self.time = 0
               self.load_json()
               self.key_categories = self.load_json()
               
               
               
               self.showNapari()
               
        def load_json(self):
            with open(self.categories_json, 'r') as f:
                return json.load(f)      
            
            
        
        def showNapari(self):
                 
                 
                 Raw_path = os.path.join(self.imagedir, self.fileextension)
                 X = glob.glob(Raw_path)
                 Imageids = []
                 
                 for imagename in X:
                     Imageids.append(imagename)
                 
                 
                 eventidbox = QComboBox()
                 eventidbox.addItem(EventBoxname)
                 for (event_name,event_label) in self.key_categories.items():
                     
                     eventidbox.addItem(event_name)
                    
                 imageidbox = QComboBox()   
                 imageidbox.addItem(Boxname)   
                 detectionsavebutton = QPushButton('Save Clicks')
                 
                 for i in range(0, len(Imageids)):
                     
                     
                     imageidbox.addItem(str(Imageids[i]))
                     
                     
                 self.figure = plt.figure(figsize=(4, 4))
                 self.multiplot_widget = FigureCanvas(self.figure)
                 self.ax = self.multiplot_widget.figure.subplots(1, 1)
                 width = 400
                 dock_widget = self.viewer.window.add_dock_widget(
                 self.multiplot_widget, name="EventStats", area='right')
                 self.multiplot_widget.figure.tight_layout()
                 self.viewer.window._qt_window.resizeDocks([dock_widget], [width], Qt.Horizontal)   
                 
                 
                 eventidbox.currentIndexChanged.connect(lambda eventid = eventidbox : self.csv_add(
                         
                         
                         os.path.basename(os.path.splitext(imageidbox.currentText())[0]),
                         eventidbox.currentText()
                    
                )
            )    
                 
                 imageidbox.currentIndexChanged.connect(
                 lambda trackid = imageidbox: self.image_add(
                         
                         imageidbox.currentText(),
                         
                         os.path.basename(os.path.splitext(imageidbox.currentText())[0])
                    
                )
            )            
                 
                    
                 self.viewer.window.add_dock_widget(imageidbox, name="Image", area='left') 
                 self.viewer.window.add_dock_widget(eventidbox, name="Event", area='left')  
                 
                 
                        
        def csv_add(self, imagename, csv_event_name ):
            
            
            self.event_name = csv_event_name
            animation_widget = AnimationWidget(self.viewer, self.savedir, imagename + csv_event_name, 0, self.image.shape[0])
            self.viewer.window.add_dock_widget(animation_widget, area='right')
            self.viewer.update_console({'animation': animation_widget.animation})
            
            for (event_name,event_label) in self.key_categories.items():
                                
                                if event_label > 0 and csv_event_name == event_name:
                                     self.event_label = event_label                         
                                     for layer in list(self.viewer.layers):
                                          
                                         if 'Detections'  in layer.name or layer.name in 'Detections' :
                                                    self.viewer.layers.remove(layer)           
                                       
                                     
                                     self.csvname = self.savedir + "/" + event_name + "Location" + (os.path.splitext(os.path.basename(imagename))[0] + '.csv')
                                     
            self.dataset   = pd.read_csv(self.csvname, delimiter = ',')
            self.dataset_index = self.dataset.index
            self.ax.cla()
            #Data is written as T, Y, X, Score, Size, Confidence
            self.T = self.dataset[self.dataset.keys()[0]][0:]
            self.Z = self.dataset[self.dataset.keys()[1]][0:]
            self.Y = self.dataset[self.dataset.keys()[2]][0:]
            self.X = self.dataset[self.dataset.keys()[3]][0:]
            self.Score = self.dataset[self.dataset.keys()[4]][0:]
            self.Size = self.dataset[self.dataset.keys()[5]][0:]
            self.Confidence = self.dataset[self.dataset.keys()[6]][0:]
            
            
            
            timelist = []
            eventlist= []
            for i in range(0, self.image.shape[0]):
                
                   currentT   = np.round(self.dataset["T"]).astype('int')
                   currentZ = np.round(self.dataset["Z"]).astype('int')
                   currentScore = self.dataset["Score"]
                   currentConf = self.dataset["Confidence"]
                   condition = currentT == i
                   condition_indices = self.dataset_index[condition]
                   conditionScore = currentScore[condition_indices]
                
                   score_condition = conditionScore > self.event_threshold[self.event_label]
                
                   countT = len(conditionScore[score_condition])
                   timelist.append(i)
                   eventlist.append(countT)
                
                
            self.ax.plot(timelist, eventlist, '-r')
            self.ax.set_title(self.event_name + "Events")
            self.ax.set_xlabel("Time")
            self.ax.set_ylabel("Counts")
            self.figure.canvas.draw()
            self.figure.canvas.flush_events()
            plt.savefig(self.savedir  + self.event_name   + '.png')        

            listtime = self.T.tolist()
            listz = self.Z.tolist()
            listy = self.Y.tolist()
            listx = self.X.tolist()
            
            listsize = self.Size.tolist()
            listscore = self.Score.tolist()
            listconfidence = self.Confidence.tolist()
            event_locations = []
            size_locations = []
            score_locations = []
            confidence_locations = []
            ndim = len(self.image)
            for i in (range(len(listtime))):
                 
                 tcenter = int(listtime[i])
                 zcenter = listz[i]
                 ycenter = listy[i]
                 xcenter = listx[i]
                 size = listsize[i]
                 score = listscore[i]
                 confidence = listconfidence[i]   
                 if score > self.event_threshold[self.event_label]:
                         event_locations.append([int(tcenter), int(ycenter), int(xcenter)])   
                         size_locations.append(size)
                         score_locations.append(score)
                         confidence_locations.append(confidence)
            point_properties = {'score' : np.array(score_locations), 'confidence' : np.array(confidence_locations)}    
            text_properties = {
            'text': self.event_name +': {score:.5f}' + '\n' + 'Confidence' +  ': {confidence:.5f}',
            'anchor': 'upper_left',
            'translation': [-5, 0],
            'size': 12,
            'color': 'pink',
        }
            for layer in list(self.viewer.layers):
                              
                             if 'Detections'  in layer.name or layer.name in 'Detections' :
                                        self.viewer.layers.remove(layer) 
            if len(score_locations) > 0:                             
                   self.viewer.add_points(event_locations, size = size_locations , properties = point_properties, text = text_properties,  name = 'Detections' + self.event_name, face_color = [0]*4, edge_color = "red", edge_width = 4) 
                   

                                     
                                        
            
        def image_add(self, image_toread, imagename):
                                    
                for layer in list(self.viewer.layers):
                                         if 'Image' in layer.name or layer.name in 'Image':
                                                    self.viewer.layers.remove(layer)
                                                    
                                                    
                self.image = imread(image_toread)
                print(self.image.shape)
                
                if self.heatmapimagedir is not None:
                     self.heat_image = imread(self.heatmapimagedir + imagename + self.heatname + '.tif')
                     self.heat_image =  MidSlices(self.heat_image, self.start_project_mid, self.end_project_mid, axis = 1)
                if len(self.image.shape) == 4:
                    self.image =  MidSlices(self.image, self.start_project_mid, self.end_project_mid, axis = 1)
                    
                
                self.viewer.add_image(self.image, name= 'Image' + imagename )
                if self.heatmapimagedir is not None:
                     self.viewer.add_image(self.heat_image, name= 'Image' + imagename + self.heatname )
                                                    
def MidSlices(Image, start_project_mid, end_project_mid, axis = 1):
    
    
    SmallImage = Image.take(indices = range(Image.shape[axis]//2 - start_project_mid, Image.shape[axis]//2 + end_project_mid), axis = axis)
    
    MaxProject = np.amax(SmallImage, axis = axis)
        
    return MaxProject  

def TruePositives(csv_gt, csv_pred, thresholdscore = 1 -  1.0E-6,  thresholdspace = 10, thresholdtime = 2):
    
            
            try:
                
                    tp = 0
                  

                    dataset_pred  = pd.read_csv(csv_pred, delimiter = ',')
                    dataset_pred_index = dataset_pred.index

                    T_pred = dataset_pred[dataset_pred.keys()[0]][0:]
                    Y_pred = dataset_pred[dataset_pred.keys()[1]][0:]
                    X_pred = dataset_pred[dataset_pred.keys()[2]][0:]
                    Score_pred = dataset_pred[dataset_pred.keys()[3]][0:]
                    
                    
                    listtime_pred = T_pred.tolist()
                    listy_pred = Y_pred.tolist()
                    listx_pred = X_pred.tolist()
                    listscore_pred = Score_pred.tolist()
                    location_pred = []
                    for i in range(len(listtime_pred)):

                        if listscore_pred[i] > thresholdscore:   
                            location_pred.append([listtime_pred[i], listy_pred[i], listx_pred[i]])

                    tree = spatial.cKDTree(location_pred)


                    dataset_gt  = pd.read_csv(csv_gt, delimiter = ',')
                    dataset_gt_index = dataset_gt.index

                    T_gt = dataset_gt[dataset_gt.keys()[0]][0:]
                    Y_gt = dataset_gt[dataset_gt.keys()[1]][0:]
                    X_gt = dataset_gt[dataset_gt.keys()[2]][0:]

                    listtime_gt = T_gt.tolist()
                    listy_gt = Y_gt.tolist()
                    listx_gt = X_gt.tolist()
                    location_gt = []
                    for i in range(len(listtime_gt)):
                        
                        index = [float(listtime_gt[i]), float(listy_gt[i]), float(listx_gt[i])]
                        closestpoint = tree.query(index)
                        spacedistance, timedistance = TimedDistance(index, location_pred[closestpoint[1]])
                        
                        if spacedistance < thresholdspace and timedistance < thresholdtime:
                            tp  = tp + 1
                    
                    fn = FalseNegatives(csv_pred, csv_gt, thresholdscore = thresholdscore, thresholdspace = thresholdspace, thresholdtime = thresholdtime)
                    fp = FalsePositives(csv_pred, csv_gt, thresholdscore = thresholdscore, thresholdspace = thresholdspace, thresholdtime = thresholdtime)
                    return tp/len(listtime_gt) * 100, fn, fp
                
            except:
                 
                 return 'File not found'
                 pass

 
def DownsampleData(image, DownsampleFactor):
                    
                    if DownsampleFactor!=1:  
                  
                        print('Downsampling Image in XY by', DownsampleFactor)
                        scale_percent = int(100/DownsampleFactor) # percent of original size
                        width = int(image.shape[2] * scale_percent / 100)
                        height = int(image.shape[1] * scale_percent / 100)
                        dim = (width, height)
                        smallimage = np.zeros([image.shape[0],  height,width])
                        for i in range(0, image.shape[0]):
                              # resize image
                              smallimage[i,:] = cv2.resize(image[i,:].astype('float32'), dim)         

                        return smallimage
                    else:

                        return image
                
def PatchGenerator(image,resultsdir,csv_gt,number_patches, patch_shape, size_tminus,size_tplus,DownsampleFactor = 1 ):
    
    
                    image = DownsampleData(image, DownsampleFactor)
                    dataset_gt  = pd.read_csv(csv_gt, delimiter = ',')
            
                    dataset_gt = dataset_gt.sample(frac = 1)
                    dataset_gt_index = dataset_gt.index
                    T_gt = dataset_gt[dataset_gt.keys()[0]][0:]
                    Y_gt = dataset_gt[dataset_gt.keys()[1]][0:]/DownsampleFactor
                    X_gt = dataset_gt[dataset_gt.keys()[2]][0:]/DownsampleFactor

                    listtime_gt = T_gt.tolist()
                    
                    listy_gt = Y_gt.tolist()
                    listx_gt = X_gt.tolist()
                    location_gt = []
                    fn = len(listtime_gt)
                    count = 0
                    Data = []
                    for i in range(len(listtime_gt)):
                        if count >  2 * number_patches:
                            break
                        time = int(float(listtime_gt[i])) - 1
                        y = float(listy_gt[i])
                        x = float(listx_gt[i])
                        
                        if x > 0.25 * image.shape[2] and x < 0.75* image.shape[2] and y > 0.25 * image.shape[1] and y < 0.75* image.shape[1]:
                                crop_Xminus = x - int(patch_shape[0] / 2)
                                crop_Xplus = x + int(patch_shape[0] / 2)
                                crop_Yminus = y - int(patch_shape[1] / 2)
                                crop_Yplus = y + int(patch_shape[1] / 2)

                          
                                randomy = np.random.randint(min(0.25 * image.shape[2],0.25 * image.shape[1]), high=max(0.25 * image.shape[2],0.25 * image.shape[1]))
                                randomx = np.random.randint(min(0.25 * image.shape[2],0.25 * image.shape[1]), high=max(0.25 * image.shape[2],0.25 * image.shape[1]))
                                random_crop_Xminus = randomx - int(patch_shape[0] / 2)
                                random_crop_Xplus = randomx + int(patch_shape[0] / 2)
                                random_crop_Yminus = randomy - int(patch_shape[1] / 2)
                                random_crop_Yplus = randomy + int(patch_shape[1] / 2)

                                region = (slice(int(time - size_tminus),int(time + size_tplus  + 1)),slice(int(crop_Yminus), int(crop_Yplus)),
                                          slice(int(crop_Xminus), int(crop_Xplus)))

                                random_region = (slice(int(time - size_tminus),int(time + size_tplus  + 1)),slice(int(random_crop_Yminus), int(random_crop_Yplus)),
                                          slice(int(random_crop_Xminus), int(random_crop_Xplus)))


                                crop_image = image[region] 
                                random_crop_image = image[random_region]
                                if(crop_image.shape[0] == size_tplus + size_tminus + 1 and crop_image.shape[1]== patch_shape[1] and crop_image.shape[2]== patch_shape[0]):
                                      Data.append([time, y * DownsampleFactor, x * DownsampleFactor])
                                      imwrite(resultsdir + 'Skeletor' + 'T' +  str(time) + 'Y' + str(y*DownsampleFactor) + 'X' + str(x*DownsampleFactor) + '.tif', crop_image.astype('float16'),metadata={'axes': 'TYX'})                
                                count = count + 1        
                                if(random_crop_image.shape[0] == size_tplus + size_tminus + 1 and random_crop_image.shape[1]== patch_shape[1] and random_crop_image.shape[2]== patch_shape[0]):
                                      Data.append([time, randomy * DownsampleFactor, randomx * DownsampleFactor])
                                      imwrite(resultsdir + 'Skeletor' + 'T' + str(time) + 'Y' + str(randomy*DownsampleFactor) + 'X' + str(randomx*DownsampleFactor) + '.tif', random_crop_image.astype('float16'),metadata={'axes': 'TYX'})
                                count = count + 1 
                    
                    writer = csv.writer(open(resultsdir + '/' + ('GTLocator') + ".csv", "w"))
                    writer.writerows(Data)
                    
                    
def FalseNegatives(csv_pred, csv_gt, thresholdscore = 1 -  1.0E-6, thresholdspace = 10, thresholdtime = 2):
    
            
            try:
                
                    
                  

                    dataset_pred  = pd.read_csv(csv_pred, delimiter = ',')
                    dataset_pred_index = dataset_pred.index

                    T_pred = dataset_pred[dataset_pred.keys()[0]][0:]
                    Y_pred = dataset_pred[dataset_pred.keys()[1]][0:]
                    X_pred = dataset_pred[dataset_pred.keys()[2]][0:]
                    Score_pred = dataset_pred[dataset_pred.keys()[3]][0:]
                    
                    listtime_pred = T_pred.tolist()
                    listy_pred = Y_pred.tolist()
                    listx_pred = X_pred.tolist()
                    listscore_pred = Score_pred.tolist()
                    location_pred = []
                    for i in range(len(listtime_pred)):
                        
                        
                        if listscore_pred[i] > thresholdscore:
                           location_pred.append([listtime_pred[i], listy_pred[i], listx_pred[i]])

                    tree = spatial.cKDTree(location_pred)


                    dataset_gt  = pd.read_csv(csv_gt, delimiter = ',')
                    dataset_gt_index = dataset_gt.index

                    T_gt = dataset_gt[dataset_gt.keys()[0]][0:]
                    Y_gt = dataset_gt[dataset_gt.keys()[1]][0:]
                    X_gt = dataset_gt[dataset_gt.keys()[2]][0:]

                    listtime_gt = T_gt.tolist()
                    listy_gt = Y_gt.tolist()
                    listx_gt = X_gt.tolist()
                    location_gt = []
                    fn = len(listtime_gt)
                    for i in range(len(listtime_gt)):
                        
                        index = [float(listtime_gt[i]), float(listy_gt[i]), float(listx_gt[i])]
                        closestpoint = tree.query(index)
                        spacedistance, timedistance = TimedDistance(index, location_pred[closestpoint[1]])

                        if spacedistance < thresholdspace and timedistance < thresholdtime:
                            fn  = fn - 1

                            


                    return fn/len(listtime_gt) * 100
                
            except:
                 
                 return 'File not found'
                 pass             
                
def FalsePositives(csv_pred, csv_gt, thresholdscore = 1 -  1.0E-6, thresholdspace = 10, thresholdtime = 2):
    
            
            try:
                
                    
                  

                    dataset_pred  = pd.read_csv(csv_pred, delimiter = ',')
                    dataset_pred_index = dataset_pred.index

                    T_pred = dataset_pred[dataset_pred.keys()[0]][0:]
                    Y_pred = dataset_pred[dataset_pred.keys()[1]][0:]
                    X_pred = dataset_pred[dataset_pred.keys()[2]][0:]
                    Score_pred = dataset_pred[dataset_pred.keys()[3]][0:]
                    
                    listtime_pred = T_pred.tolist()
                    listy_pred = Y_pred.tolist()
                    listx_pred = X_pred.tolist()
                    listscore_pred = Score_pred.tolist()
                    location_pred = []
                    


                    dataset_gt  = pd.read_csv(csv_gt, delimiter = ',')
                    dataset_gt_index = dataset_gt.index

                    T_gt = dataset_gt[dataset_gt.keys()[0]][0:]
                    Y_gt = dataset_gt[dataset_gt.keys()[1]][0:]
                    X_gt = dataset_gt[dataset_gt.keys()[2]][0:]

                    listtime_gt = T_gt.tolist()
                    listy_gt = Y_gt.tolist()
                    listx_gt = X_gt.tolist()
                    location_gt = []
                    fp = len(listtime_pred)
                    
                    for i in range(len(listtime_gt)):
                        
                     
                           location_gt.append([listtime_gt[i], listy_gt[i], listx_gt[i]])

                    tree = spatial.cKDTree(location_gt)
                    for i in range(len(listtime_pred)):
                        
                        index = [float(listtime_pred[i]), float(listy_pred[i]), float(listx_pred[i])]
                        closestpoint = tree.query(index)
                        spacedistance, timedistance = TimedDistance(index, location_gt[closestpoint[1]])

                        if spacedistance < thresholdspace and timedistance < thresholdtime:
                            fp  = fp - 1

                            


                    return fp/len(listtime_pred) * 100
                
            except:
                 
                 return 'File not found'
                 pass             
                                
 
def TimedDistance(pointA, pointB):

    
     spacedistance = float(np.sqrt( (pointA[1] - pointB[1] ) * (pointA[1] - pointB[1] ) + (pointA[2] - pointB[2] ) * (pointA[2] - pointB[2] )  ))
     
     timedistance = float(np.abs(pointA[0] - pointB[0]))
     
     
     return spacedistance, timedistance
                
                
def GetMarkers(image):
    
    
    MarkerImage = np.zeros(image.shape)
    waterproperties = measure.regionprops(image)                
    Coordinates = [prop.centroid for prop in waterproperties]
    Coordinates = sorted(Coordinates , key=lambda k: [k[0], k[1]])
    MarkerImage[tuple(coordinates_int.T)] = 1 + np.arange(len(Coordinates))

    markers = morphology.dilation(MarkerImage, morphology.disk(2))        
   
    return markers 


"""
The easing functions here were adapted from
https://gist.github.com/zeffii/c1e14dd6620ad855d81ec2e89a859719
original c code:
https://raw.githubusercontent.com/warrenm/AHEasing/master/AHEasing/easing.c
Copyright (c) 2011, Auerhaus Development, LLC
http://sam.zoy.org/wtfpl/COPYING for more details.
"""
from enum import Enum
from functools import partial
from math import cos, pi, pow, sin, sqrt

tau = pi * 2


def linear_interpolation(p):
    """Modeled after the line y = x"""
    return p


def quadratic_ease_in(p):
    """Modeled after the parabola y = x^2"""
    return p * p


def quadratic_ease_out(p):
    """Modeled after the parabola y = -x^2 + 2x"""
    return -(p * (p - 2))


def quadratic_ease_in_out(p):
    """Modeled after the piecewise quadratic
    y = (1/2)((2x)^2)             ; [0, 0.5)
    y = -(1/2)((2x-1)*(2x-3) - 1) ; [0.5, 1]
    """
    if p < 0.5:
        return 2 * p * p
    else:
        return (-2 * p * p) + (4 * p) - 1


def cubic_ease_in(p):
    """Modeled after the cubic y = x^3"""
    return p * p * p


def cubic_ease_out(p):
    """Modeled after the cubic y = (x - 1)^3 + 1"""
    f = p - 1
    return (f * f * f) + 1


def cubic_ease_in_out(p):
    """Modeled after the piecewise cubic
    y = (1/2)((2x)^3)       ; [0, 0.5)
    y = (1/2)((2x-2)^3 + 2) ; [0.5, 1]
    """
    if p < 0.5:
        return 4 * p * p * p
    else:
        f = (2 * p) - 2
        return (0.5 * f * f * f) + 1


def quintic_ease_in(p):
    """Modeled after the quintic y = x^5"""
    return p * p * p * p * p


def quintic_ease_out(p):
    """Modeled after the quintic y = (x - 1)^5 + 1"""
    f = p - 1
    return (f * f * f * f * f) + 1


def quintic_ease_in_out(p):
    """Modeled after the piecewise quintic
    y = (1/2)((2x)^5)       ; [0, 0.5)
    y = (1/2)((2x-2)^5 + 2) ; [0.5, 1]
    """
    if p < 0.5:
        return 16 * p * p * p * p * p
    else:
        f = (2 * p) - 2
        return (0.5 * f * f * f * f * f) + 1


def sine_ease_in(p):
    """Modeled after quarter-cycle of sine wave"""
    return sin((p - 1) * tau) + 1


def sine_ease_out(p):
    """Modeled after quarter-cycle of sine wave (different phase)"""
    return sin(p * tau)


def sine_ease_in_out(p):
    """Modeled after half sine wave"""
    return 0.5 * (1 - cos(p * pi))


def circular_ease_in(p):
    """Modeled after shifted quadrant IV of unit circle"""
    return 1 - sqrt(1 - (p * p))


def circular_ease_out(p):
    """Modeled after shifted quadrant II of unit circle"""
    return sqrt((2 - p) * p)


def circular_ease_in_out(p):
    """Modeled after the piecewise circular function
    y = (1/2)(1 - sqrt(1 - 4x^2))           ; [0, 0.5)
    y = (1/2)(sqrt(-(2x - 3)*(2x - 1)) + 1) ; [0.5, 1]
    """
    if p < 0.5:
        return 0.5 * (1 - sqrt(1 - 4 * (p * p)))
    else:
        return 0.5 * (sqrt(-((2 * p) - 3) * ((2 * p) - 1)) + 1)


def exponential_ease_in(p):
    """Modeled after the exponential function y = 2^(10(x - 1))"""
    if p == 0.0:
        return p
    else:
        return pow(2, 10 * (p - 1))


def exponential_ease_out(p):
    """Modeled after the exponential function y = -2^(-10x) + 1"""
    if p == 1.0:
        return p
    else:
        return 1 - pow(2, -10 * p)


def exponential_ease_in_out(p):
    """Modeled after the piecewise exponential
    y = (1/2)2^(10(2x - 1))         ; [0,0.5)
    y = -(1/2)*2^(-10(2x - 1))) + 1 ; [0.5,1]
    """
    if p == 0.0 or p == 1.0:
        return p

    if p < 0.5:
        return 0.5 * pow(2, (20 * p) - 10)
    else:
        return -0.5 * pow(2, (-20 * p) + 10) + 1


def elastic_ease_in(p):
    """Modeled after the damped sine wave y = sin(13pi/2*x)*2^(10 * (x - 1))"""
    return sin(13 * tau * p) * pow(2, 10 * (p - 1))


def elastic_ease_out(p):
    """Modeled after the damped sine wave y = sin(-13pi/2*(x + 1))*pow(2, -10x) + 1"""
    return sin(-13 * tau * (p + 1)) * pow(2, -10 * p) + 1


def elastic_ease_in_out(p):
    """Modeled after the piecewise exponentially-damped sine wave:
    y = (1/2)*sin(13pi/2*(2*x))*pow(2, 10 * ((2*x) - 1))      ; [0, 0.5)
    y = (1/2)*(sin(-13pi/2*((2x-1)+1))*pow(2,-10(2*x-1)) + 2) ; [0.5, 1]
    """
    if p < 0.5:
        return 0.5 * sin(13 * tau * (2 * p)) * pow(2, 10 * ((2 * p) - 1))
    else:
        return 0.5 * (
            sin(-13 * tau * ((2 * p - 1) + 1)) * pow(2, -10 * (2 * p - 1)) + 2
        )


def back_ease_in(p):
    """Modeled after the overshooting cubic y = x^3-x*sin(x*pi)"""
    return p * p * p - p * sin(p * pi)


def back_ease_out(p):
    """Modeled after overshooting cubic y = 1-((1-x)^3-(1-x)*sin((1-x)*pi))"""
    f = 1 - p
    return 1 - (f * f * f - f * sin(f * pi))


def back_ease_in_out(p):
    """Modeled after the piecewise overshooting cubic function:
    y = (1/2)*((2x)^3-(2x)*sin(2*x*pi))           ; [0, 0.5)
    y = (1/2)*(1-((1-x)^3-(1-x)*sin((1-x)*pi))+1) ; [0.5, 1]
    """
    if p < 0.5:
        f = 2 * p
        return 0.5 * (f * f * f - f * sin(f * pi))
    else:
        f = 1 - (2 * p - 1)
        return (0.5 * (1 - (f * f * f - f * sin(f * pi)))) + 0.5


def bounce_ease_in(p):
    return 1 - bounce_ease_out(1 - p)


def bounce_ease_out(p):
    if p < 4 / 11.0:
        return (121 * p * p) / 16.0

    elif p < 8 / 11.0:
        return ((363 / 40.0) * p * p) - ((99 / 10.0) * p) + (17 / 5.0)

    elif p < 9 / 10.0:
        return ((4356 / 361.0) * p * p) - ((35442 / 1805.0) * p) + (16061 / 1805.0)

    else:
        return ((54 / 5.0) * p * p) - ((513 / 25.0) * p) + (268 / 25.0)


def bounce_ease_in_out(p):
    if p < 0.5:
        return 0.5 * bounce_ease_in(p * 2)
    else:
        return (0.5 * bounce_ease_out(p * 2 - 1)) + 0.5


class Easing(Enum):
    """Easing: easing function to use for a transition.

    Selects a preset easing function
            * linear: linear interpolation between start and endpoint.
            * quadratic: quadratic easing in and out.
                Modeled after the piecewise quadratic
                y = (1/2)((2x)^2)             ; [0, 0.5)
                y = -(1/2)((2x-1)*(2x-3) - 1) ; [0.5, 1]
            * cubic: cubic easing in and out.
                Modeled after the piecewise cubic
                y = (1/2)((2x)^3)       ; [0, 0.5)
                y = (1/2)((2x-2)^3 + 2) ; [0.5, 1]
            * quintic: quintic easing in and out.
                Modeled after the piecewise quintic
                y = (1/2)((2x)^5)       ; [0, 0.5)
                y = (1/2)((2x-2)^5 + 2) ; [0.5, 1]
            * sine: sinusoidal easing in and out.
                Modeled after half sine wave
                y = 0.5 * (1 - cos(x * pi))
            * circular: circular easing in and out.
                Modeled after the piecewise circular function
                y = (1/2)(1 - sqrt(1 - 4x^2))           ; [0, 0.5)
                y = (1/2)(sqrt(-(2x - 3)*(2x - 1)) + 1) ; [0.5, 1]
            * exponential: exponential easing in and out.
                Modeled after the piecewise exponential
                y = (1/2)2^(10(2x - 1))         ; [0,0.5)
                y = -(1/2)*2^(-10(2x - 1))) + 1 ; [0.5,1]
            * elastic: elastic easing in and out.
                Modeled after the piecewise exponentially-damped sine wave:
                y = (1/2)*sin(13pi/2*(2*x))*pow(2, 10 * ((2*x) - 1))      ; [0, 0.5)
                y = (1/2)*(sin(-13pi/2*((2x-1)+1))*pow(2,-10(2*x-1)) + 2) ; [0.5, 1]
            * back: back easing in and out.
                Modeled after the piecewise overshooting cubic function:
                y = (1/2)*((2x)^3-(2x)*sin(2*x*pi))           ; [0, 0.5)
                y = (1/2)*(1-((1-x)^3-(1-x)*sin((1-x)*pi))+1) ; [0.5, 1]
            * bounce: bounce easing in and out.
    """

    LINEAR = partial(linear_interpolation)
    QUADRATIC = partial(quadratic_ease_in_out)
    CUBIC = partial(cubic_ease_in_out)
    QUINTIC = partial(quintic_ease_in_out)
    SINE = partial(sine_ease_in_out)
    CIRCULAR = partial(circular_ease_in_out)
    EXPONENTIAL = partial(exponential_ease_in_out)
    ELASTIC = partial(elastic_ease_in_out)
    BACK = partial(back_ease_in_out)
    BOUNCE = partial(bounce_ease_in_out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 15:53:01 2020

@author: Varun Kapoor
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 15:25:10 2020

@author: aimachine
"""
import argparse
import numpy as np
class static_config(argparse.Namespace):
    
    def __init__(self, npz_directory = None, npz_name = None, npz_val_name = None, key_categories = None, key_cord = None, residual = True,stage_number = 3, last_conv_factor = 4,  stride = 4, gridx = 1, gridy = 1, nboxes = 1, 
                 depth = 29, start_kernel = 3, mid_kernel = 3,startfilter = 32, show = True, imagex = 64, imagey = 64,
                 epochs =100,  learning_rate = 1.0E-4, batch_size = 10, model_name = 'NEATModel', yolo_v0 = True, multievent = True,  **kwargs):
        


           self.npz_directory = npz_directory
           self.npz_name = npz_name
           self.npz_val_name = npz_val_name
           self.key_categories = key_categories
           self.key_cord = key_cord
           self.residual = residual
           self.depth = depth
           self.start_kernel = start_kernel
           self.mid_kernel = mid_kernel
           self.startfilter = startfilter
           self.gridx = gridx
           self.gridy = gridy
           self.nboxes = nboxes
           self.show = show
           self.imagex = imagex
           self.imagey = imagey
           self.epochs = epochs
           self.stride = stride
           self.stage_number = stage_number
           self.last_conv_factor = last_conv_factor
           self.yolo_v0 = yolo_v0
           self.categories = len(self.key_categories)
           self.box_vector = len(self.key_cord)
           self.learning_rate = learning_rate
           self.batch_size = batch_size
           self.model_name = model_name
           self.multievent = multievent
          
           self.is_valid()      
           
    def to_json(self):

         config = {
             
                 'npz_directory' : self.npz_directory,
                 'npz_name' : self.npz_name,
                 'npz_val_name' : self.npz_val_name,
                 'residual' : self.residual,
                 'depth' : self.depth,
                 'start_kernel' : self.start_kernel,
                 'mid_kernel' : self.mid_kernel,
                 'startfilter' : self.startfilter,
                 'gridx' : self.gridx,
                 'gridy' : self.gridy,
                 'nboxes' : self.nboxes,
                 'imagex' : self.imagex,
                 'imagey' : self.imagey,
                 'stride' : self.stride,
                 'epochs' : self.epochs,
                 'categories' : self.categories,
                 'box_vector' : self.box_vector,
                 'learning_rate' : self.learning_rate,
                 'batch_size' : self.batch_size,
                 'model_name' : self.model_name,
                 'multievent' : self.multievent,
                 'yolo_v0': self.yolo_v0,
                 'show' : self.show,
                 'stage_number' : self.stage_number,
                 'last_conv_factor' : self.last_conv_factor
                 
                 }
         for (k,v) in self.key_categories.items():
             config[k] = v
             
         for (k,v) in self.key_cord.items():
             config[k] = v    
         
         
         
         return config
                
    

              
    def is_valid(self, return_invalid=False):
            """Check if configuration is valid.
            Returns
            -------
            bool
            Flag that indicates whether the current configuration values are valid.
            """
            def _is_int(v,low=None,high=None):
              return (
                isinstance(v,int) and
                (True if low is None else low <= v) and
                (True if high is None else v <= high)
              )

            ok = {}
            ok['npz_directory'] = isinstance(self.npz_directory, str)
            ok['npz_name'] = isinstance(self.npz_name, str)
            ok['npz_val_name'] = isinstance(self.npz_val_name, str)
            ok['residual'] = isinstance(self.residual,bool)
            ok['yolo_v0'] = isinstance(self.yolo_v0,bool)
            ok['depth']         = _is_int(self.depth,1)
            ok['stride']         = _is_int(self.stride,1)
            ok['start_kernel']       = _is_int(self.start_kernel,1)
            ok['mid_kernel']         = _is_int(self.mid_kernel,1)
            ok['startfilter']        = _is_int(self.startfilter, 1)
            ok['stage_number']        = _is_int(self.stage_number, 1)
            ok['last_conv_factor']        = _is_int(self.last_conv_factor, 1)
            ok['epochs']        = _is_int(self.epochs, 1)
            ok['nboxes']       = _is_int(self.nboxes, 1)
            ok['gridx'] = _is_int(self.gridx, 1)
            ok['gridy'] = _is_int(self.gridy, 1)
            ok['imagex']        = _is_int(self.imagex, 1)
            ok['imagey']         = _is_int(self.imagey,1)
            ok['show'] = isinstance(self.show,bool)
            ok['learning_rate'] = np.isscalar(self.learning_rate) and self.learning_rate > 0
            ok['multievent'] = isinstance(self.multievent,bool)
            ok['categories'] =  _is_int(len(self.key_categories), 1)
            ok['box_vector'] =  _is_int(self.box_vector, 1)
            
            
            if return_invalid:
                return all(ok.values()), tuple(k for (k,v) in ok.items() if not v)
            else:
                return all(ok.values())
                   

        


from oneat.NEATUtils import plotters
import numpy as np
from oneat.NEATUtils import helpers
from oneat.NEATUtils.helpers import get_nearest,  load_json, yoloprediction, normalizeFloatZeroOne, GenerateMarkers, Generate_only_mask, MakeTrees, DownsampleData,save_dynamic_csv, dynamic_nms, gold_nms
from keras import callbacks
import os
from scipy.ndimage.morphology import binary_dilation, binary_erosion
import math
from tqdm import tqdm
from oneat.NEATModels import nets
from oneat.NEATModels.nets import Concat
from oneat.NEATModels.loss import dynamic_yolo_loss
from keras import backend as K
import tensorflow as tf
from tensorflow.keras import optimizers
from pathlib import Path
from keras.models import load_model
from tifffile import imread, imwrite
import napari
import glob
from skimage.morphology import erosion, dilation, disk
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QComboBox, QPushButton
import cv2
from scipy import ndimage
from skimage.measure import label
from skimage import measure
Boxname = 'ImageIDBox'
EventBoxname = 'EventIDBox'


class NEATDynamic(object):
    """
    Parameters
    ----------
    
    NpzDirectory : Specify the location of npz file containing the training data with movies and labels
    
    TrainModelName : Specify the name of the npz file containing training data and labels
    
    ValidationModelName :  Specify the name of the npz file containing validation data and labels
    
    categories : Number of action classes
    
    Categories_Name : List of class names and labels
    
    model_dir : Directory location where trained model weights are to be read or written from
    
    model_name : The h5 file of CNN + LSTM + Dense Neural Network to be used for training
    
    model_keras : The model as it appears as a Keras function
    
    model_weights : If re-training model_weights = model_dir + model_name else None as default
    
    lstm_hidden_units : Number of hidden uniots for LSTm layer, 64 by default
    
    epochs :  Number of training epochs, 55 by default
    
    batch_size : batch_size to be used for training, 20 by default
    
    
    
    """

    def __init__(self, config, model_dir, model_name,  catconfig=None, cordconfig=None):

        self.config = config
        self.catconfig = catconfig
        self.cordconfig = cordconfig
        self.model_dir = model_dir
        self.model_name = model_name
        if self.config != None:
            self.npz_directory = config.npz_directory
            self.npz_name = config.npz_name
            self.npz_val_name = config.npz_val_name
            self.key_categories = config.key_categories
            self.stage_number = config.stage_number
            self.last_conv_factor = config.last_conv_factor
            self.show = config.show
            self.key_cord = config.key_cord
            self.box_vector = len(config.key_cord)
            self.categories = len(config.key_categories)
            self.depth = config.depth
            self.start_kernel = config.start_kernel
            self.mid_kernel = config.mid_kernel
            self.lstm_kernel = config.lstm_kernel
            self.learning_rate = config.learning_rate
            self.epochs = config.epochs
            self.residual = config.residual
            self.startfilter = config.startfilter
            self.batch_size = config.batch_size
            self.multievent = config.multievent
            self.imagex = config.imagex
            self.imagey = config.imagey
            self.imaget = config.size_tminus + config.size_tplus + 1
            self.size_tminus = config.size_tminus
            self.size_tplus = config.size_tplus

            self.nboxes = config.nboxes
            self.gridx = 1
            self.gridy = 1
            self.gridt = 1
            self.yolo_v0 = config.yolo_v0
            self.yolo_v1 = config.yolo_v1
            self.yolo_v2 = config.yolo_v2
            self.stride = config.stride
            self.lstm_hidden_unit = config.lstm_hidden_unit
        if self.config == None:

            try:
                self.config = load_json(self.model_dir + os.path.splitext(self.model_name)[0] + '_Parameter.json')
            except:
                self.config = load_json(self.model_dir + self.model_name + '_Parameter.json')

            self.npz_directory = self.config['npz_directory']
            self.npz_name = self.config['npz_name']
            self.npz_val_name = self.config['npz_val_name']
            self.key_categories = self.catconfig
            self.box_vector = self.config['box_vector']
            self.show = self.config['show']
            self.key_cord = self.cordconfig
            self.categories = len(self.catconfig)
            self.depth = self.config['depth']
            self.start_kernel = self.config['start_kernel']
            self.mid_kernel = self.config['mid_kernel']
            self.lstm_kernel = self.config['lstm_kernel']
            self.lstm_hidden_unit = self.config['lstm_hidden_unit']
            self.learning_rate = self.config['learning_rate']
            self.epochs = self.config['epochs']
            self.residual = self.config['residual']
            self.startfilter = self.config['startfilter']
            self.batch_size = self.config['batch_size']
            self.multievent = self.config['multievent']
            self.imagex = self.config['imagex']
            self.imagey = self.config['imagey']
            self.imaget = self.config['size_tminus'] + self.config['size_tplus'] + 1
            self.size_tminus = self.config['size_tminus']
            self.size_tplus = self.config['size_tplus']
            self.nboxes = self.config['nboxes']
            self.stage_number = self.config['stage_number']
            self.last_conv_factor = self.config['last_conv_factor']
            self.gridx = 1
            self.gridy = 1
            self.gridt = 1
            self.yolo_v0 = self.config['yolo_v0']
            self.yolo_v1 = self.config['yolo_v1']
            self.yolo_v2 = self.config['yolo_v2']
            self.stride = self.config['stride']
            self.lstm_hidden_unit = self.config['lstm_hidden_unit']

        self.X = None
        self.Y = None
        self.axes = None
        self.X_val = None
        self.Y_val = None
        self.Trainingmodel = None
        self.Xoriginal = None
        self.Xoriginal_val = None

        if self.residual:
            self.model_keras = nets.ORNET
        else:
            self.model_keras = nets.OSNET

        if self.multievent == True:
            self.last_activation = 'sigmoid'
            self.entropy = 'binary'

        if self.multievent == False:
            self.last_activation = 'softmax'
            self.entropy = 'notbinary'
        self.yololoss = dynamic_yolo_loss(self.categories, self.gridx, self.gridy, self.gridt, self.nboxes,
                                          self.box_vector, self.entropy, self.yolo_v0, self.yolo_v1, self.yolo_v2)

    def loadData(self):

        (X, Y), axes = helpers.load_full_training_data(self.npz_directory, self.npz_name, verbose=True)

        (X_val, Y_val), axes = helpers.load_full_training_data(self.npz_directory, self.npz_val_name, verbose=True)

        self.Xoriginal = X
        self.Xoriginal_val = X_val

        self.X = X
        self.Y = Y[:, :, 0]
        self.X_val = X_val
        self.Y_val = Y_val[:, :, 0]

        self.axes = axes
        self.Y = self.Y.reshape((self.Y.shape[0], 1, 1, self.Y.shape[1]))
        self.Y_val = self.Y_val.reshape((self.Y_val.shape[0], 1, 1, self.Y_val.shape[1]))

    def TrainModel(self):

        input_shape = (self.X.shape[1], self.X.shape[2], self.X.shape[3], self.X.shape[4])

        Path(self.model_dir).mkdir(exist_ok=True)

        if self.yolo_v2:

            for i in range(self.Y.shape[0]):

                if self.Y[i, :, :, 0] == 1:
                    self.Y[i, :, :, -1] = 1
            for i in range(self.Y_val.shape[0]):

                if self.Y_val[i, :, :, 0] == 1:
                    self.Y_val[i, :, :, -1] = 1
        Y_rest = self.Y[:, :, :, self.categories:]
 

        model_weights = self.model_dir + self.model_name
        if os.path.exists(model_weights):

            self.model_weights = model_weights
            print('loading weights')
        else:

            self.model_weights = None

        dummyY = np.zeros(
            [self.Y.shape[0], self.Y.shape[1], self.Y.shape[2], self.categories + self.nboxes * self.box_vector])
        dummyY[:, :, :, :self.Y.shape[3]] = self.Y

        dummyY_val = np.zeros([self.Y_val.shape[0], self.Y_val.shape[1], self.Y_val.shape[2],
                               self.categories + self.nboxes * self.box_vector])
        dummyY_val[:, :, :, :self.Y_val.shape[3]] = self.Y_val

        for b in range(1, self.nboxes):
            dummyY[:, :, :, self.categories + b * self.box_vector:self.categories + (b + 1) * self.box_vector] = self.Y[
                                                                                                                 :, :,
                                                                                                                 :,
                                                                                                                 self.categories: self.categories + self.box_vector]
            dummyY_val[:, :, :,
            self.categories + b * self.box_vector:self.categories + (b + 1) * self.box_vector] = self.Y_val[:, :, :,
                                                                                                 self.categories: self.categories + self.box_vector]

        self.Y = dummyY
        self.Y_val = dummyY_val


        self.Trainingmodel = self.model_keras(input_shape, self.categories, unit=self.lstm_hidden_unit,
                                              box_vector=Y_rest.shape[-1], nboxes=self.nboxes,
                                              stage_number=self.stage_number, last_conv_factor=self.last_conv_factor,
                                              depth=self.depth, start_kernel=self.start_kernel,
                                              mid_kernel=self.mid_kernel, lstm_kernel=self.lstm_kernel,
                                              startfilter=self.startfilter, input_weights=self.model_weights,
                                              last_activation=self.last_activation)

        sgd = optimizers.SGD(lr=self.learning_rate, momentum=0.99, decay=1e-6, nesterov=True)
        self.Trainingmodel.compile(optimizer=sgd, loss=self.yololoss, metrics=['accuracy'])

        self.Trainingmodel.summary()
        # Keras callbacks
        lrate = callbacks.ReduceLROnPlateau(monitor='loss', factor=0.1, patience=4, verbose=1)
        hrate = callbacks.History()
        srate = callbacks.ModelCheckpoint(self.model_dir + self.model_name, monitor='loss', verbose=1,
                                          save_best_only=False, save_weights_only=False, mode='auto', period=1)
        prate = plotters.PlotHistory(self.Trainingmodel, self.X_val, self.Y_val, self.key_categories, self.key_cord,
                                     self.gridx, self.gridy, plot=self.show, nboxes=self.nboxes)

        # Train the model and save as a h5 file
        self.Trainingmodel.fit(self.X, self.Y, batch_size=self.batch_size,
                               epochs=self.epochs, validation_data=(self.X_val, self.Y_val), shuffle=True,
                               callbacks=[lrate, hrate, srate, prate])

        # Removes the old model to be replaced with new model, if old one exists
        if os.path.exists(self.model_dir + self.model_name):
            os.remove(self.model_dir + self.model_name)

        self.Trainingmodel.save(self.model_dir + self.model_name)

    def get_markers(self, imagename, starmodel, maskmodel, savedir, n_tiles, markerdir=None, maskdir = None, watersheddir = None,
     downsamplefactor = 1, remove_markers = True):

        self.starmodel = starmodel
        self.imagename = imagename
        self.image = imread(imagename)
        self.markerdir = markerdir
        self.maskdir = maskdir
        self.maskmodel = maskmodel
        self.waterhseddir = watersheddir
        Name = os.path.basename(os.path.splitext(self.imagename)[0])
        self.savedir = savedir
        Path(self.savedir).mkdir(exist_ok=True)
        self.downsamplefactor = downsamplefactor
        self.n_tiles = n_tiles
        

        print('Obtaining Markers, Mask and Watershed image')
        if self.markerdir is None:
            self.markers, self.watershed, self.mask = GenerateMarkers(self.image, self.starmodel, self.maskmodel, self.n_tiles)
            self.markerdir = self.savedir + '/' + 'Markers'
            self.watersheddir = self.savedir + '/' + 'Watershed'
            self.maskdir = self.savedir + '/' + 'Mask'
            Path(self.markerdir).mkdir(exist_ok=True)
            Path(self.watersheddir).mkdir(exist_ok=True)
            Path(self.maskdir).mkdir(exist_ok=True)
            imwrite(self.markerdir + '/' + Name + '.tif', self.markers.astype('float32'))
            imwrite(self.watersheddir + '/' + Name + '.tif', self.watershed.astype('float32'))
            imwrite(self.maskdir + '/' + Name + '.tif', self.mask.astype('float32'))


        else:
            try:
                self.markers = imread(self.markerdir + '/' + Name + '.tif')
                self.watershed = imread(self.watersheddir + '/' + Name + '.tif')
                self.mask = imread(self.maskdir + '/' + Name + '.tif')
                if remove_markers:
                    self.markers = DownsampleData(self.markers, self.downsamplefactor)
                for i in range(0, self.markers.shape[0]):
                    self.markers[i,:] = self.markers[i,:] > 0
                    self.markers[i,:] = label(self.markers[i,:].astype('uint16'))
                    


            except:
                self.markers, self.watershed, self.mask = GenerateMarkers(self.image, self.starmodel, self.maskmodel, self.n_tiles)
                self.markerdir = self.savedir + '/' + 'Markers'
                self.watersheddir = self.savedir + '/' + 'Watershed'
                self.maskdir = self.savedir + '/' + 'Mask'
                Path(self.markerdir).mkdir(exist_ok=True)
                Path(self.watersheddir).mkdir(exist_ok=True)
                Path(self.maskdir).mkdir(exist_ok=True)
                imwrite(self.markerdir + '/' + Name + '.tif', self.markers.astype('float32'))
                imwrite(self.watersheddir + '/' + Name + '.tif', self.watershed.astype('float32'))
                imwrite(self.maskdir + '/' + Name + '.tif', self.mask.astype('float32'))
        self.marker_tree = MakeTrees(self.markers)


        return self.markers, self.marker_tree, self.watershed, self.mask
    
    def predict(self, imagename,  savedir, n_tiles=(1, 1), overlap_percent=0.8,
                event_threshold=0.5, event_confidence = 0.5, iou_threshold=0.1,  fidelity=5, downsamplefactor = 1, start_project_mid = 4, end_project_mid = 4,
                maskfilter = 10, markers = None, marker_tree = None, watershed = None,  remove_markers = True,maskmodel = None, maskdir = None):

        self.watershed = watershed
        self.maskdir = maskdir
        self.imagename = imagename
        self.Name = os.path.basename(os.path.splitext(self.imagename)[0])
        if self.maskdir is not None:
            self.maskimage = imread(self.maskdir + '/' + self.Name + '.tif')
        else:
            self.maskimage = None    
        self.image = imread(imagename)
        self.start_project_mid = start_project_mid
        self.end_project_mid = end_project_mid
        self.maskmodel = maskmodel
        self.ndim = len(self.image.shape)
        self.z = 0
        if self.ndim == 4:
           self.z = self.image.shape[1]//2
           print(f'Image {self.image.shape} is {self.ndim} dimensional, projecting around the center {self.image.shape[1]//2} - {self.start_project_mid} to {self.image.shape[1]//2} + {self.end_project_mid}') 
           self.image =  MidSlices(self.image, self.start_project_mid, self.end_project_mid, axis = 1)
           
      
        self.maskfilter = maskfilter
        
        self.heatmap = np.zeros(self.image.shape, dtype = 'float32')  
        self.savedir = savedir
        self.n_tiles = n_tiles
        self.fidelity = fidelity
        self.overlap_percent = overlap_percent
        self.iou_threshold = iou_threshold
        self.event_threshold = event_threshold
        self.event_confidence = event_confidence
        self.downsamplefactor = downsamplefactor
        self.originalimage = self.image
        if self.maskmodel is not None and self.maskimage is None:

            print(f'Generating mask, hang on')
            self.maskdir = self.savedir + '/' + 'Mask'
            Path(self.maskdir).mkdir(exist_ok=True)
            self.maskimage = Generate_only_mask(self.image, self.maskmodel, self.n_tiles)
            self.maskimage = binary_erosion(self.maskimage, size = self.maskfilter)
            imwrite(self.maskdir + '/' + self.Name + '.tif', self.mask.astype('float32'))
            print(f'Mask generated and saved at {self.maskdir}')
        

        self.model = load_model(self.model_dir + self.model_name + '.h5',
                                custom_objects={'loss': self.yololoss, 'Concat': Concat})

        self.markers = markers
        self.marker_tree = marker_tree
        self.remove_markers = remove_markers
        if self.remove_markers:
             self.image = DownsampleData(self.image, self.downsamplefactor)
        
        if self.remove_markers == True:
           self.first_pass_predict()
        if self.remove_markers == False:  
           self.second_pass_predict()
        if self.remove_markers == None:
           self.default_pass_predict() 

    def default_pass_predict(self):
        eventboxes = []
        classedboxes = {}    
        count = 0
        heatsavename = self.savedir+ "/"  + (os.path.splitext(os.path.basename(self.imagename))[0])+ '_Heat' 

        print('Detecting event locations')
        self.image = DownsampleData(self.image, self.downsamplefactor)
        for inputtime in tqdm(range(0, self.image.shape[0])):
                    if inputtime < self.image.shape[0] - self.imaget:
                                count = count + 1
                                if inputtime%(self.image.shape[0]//4)==0 and inputtime > 0 or inputtime >= self.image.shape[0] - self.imaget - 1:
                                      
                                                                              
                                      
                                      imwrite((heatsavename + '.tif' ), self.heatmap)
                                      
                                smallimage = CreateVolume(self.image, self.imaget, inputtime)

                                smallimage = normalizeFloatZeroOne(smallimage,1,99.8)
                                # Cut off the region for training movie creation
                                #Break image into tiles if neccessary
                                predictions, allx, ally = self.predict_main(smallimage)
                                #Iterate over tiles
                                for p in range(0,len(predictions)):   
                        
                                  sum_time_prediction = predictions[p]
                                  if sum_time_prediction is not None:
                                     #For each tile the prediction vector has shape N H W Categories + Training Vector labels
                                     for i in range(0, sum_time_prediction.shape[0]):
                                          time_prediction =  sum_time_prediction[i]
                                          boxprediction = yoloprediction(ally[p], allx[p], time_prediction, self.stride, inputtime , self.config, self.key_categories, self.key_cord, self.nboxes, 'detection', 'dynamic',marker_tree=self.marker_tree)
                                          
                                          if boxprediction is not None:
                                                  eventboxes = eventboxes + boxprediction
                                            
                                for (event_name,event_label) in self.key_categories.items(): 
                                                     
                                                if event_label > 0:
                                                     current_event_box = []
                                                     for box in eventboxes:
                                                
                                                        event_prob = box[event_name]
                                                        event_confidence = box['confidence']
                                                        if event_prob >= self.event_threshold and event_confidence >= self.event_confidence:
                                                           
                                                            current_event_box.append(box)
                                                     classedboxes[event_name] = [current_event_box]
                                                 
                                self.classedboxes = classedboxes    
                                self.eventboxes =  eventboxes
                                #nms over time
                                if inputtime > 0:
 
                                    self.nms()
                                    self.to_csv()
                                    eventboxes = []
                                    classedboxes = {}    
                                    count = 0


    def first_pass_predict(self):
        
        print('Detecting background event locations')
        eventboxes = []
        classedboxes = {}
        remove_candidates = {}
        
        
        for inputtime in tqdm(range(0, self.image.shape[0])):
            if inputtime < self.image.shape[0] - self.imaget:
                
                remove_candidates_list = []
                smallimage = CreateVolume(self.image, self.imaget, inputtime)
                smallimage = normalizeFloatZeroOne(smallimage, 1, 99.8)
                # Cut off the region for training movie creation
                # Break image into tiles if neccessary
                predictions, allx, ally = self.predict_main(smallimage)
                # Iterate over tiles
                for p in range(0, len(predictions)):

                    sum_time_prediction = predictions[p]

                    if sum_time_prediction is not None:
                        # For each tile the prediction vector has shape N H W Categories + Training Vector labels
                        for i in range(0, sum_time_prediction.shape[0]):
                            time_prediction = sum_time_prediction[i]
                            boxprediction = yoloprediction(ally[p], allx[p], time_prediction, self.stride,
                                                           inputtime, self.config,
                                                           self.key_categories, self.key_cord, self.nboxes, 'detection',
                                                           'dynamic', self.marker_tree, self.remove_markers)

                            if boxprediction is not None:
                                eventboxes = eventboxes + boxprediction
        for (event_name, event_label) in self.key_categories.items():
                    
                    if event_label == 0:  
                                current_event_box = []              
                                for box in eventboxes:
                
                                    event_prob = box[event_name]
                                    if event_prob >= 0.5:
                                          
                                        

                                        current_event_box.append(box)
                                        classedboxes[event_name] = [current_event_box]

        self.classedboxes = classedboxes
        if len(self.classedboxes) > 0:
                self.fast_nms()

                for box in self.iou_classedboxes:
                        ycentermean, xcentermean = get_nearest(self.marker_tree, box['ycenter'], box['xcenter'], box['real_time_event'])

                        if remove_candidates[str(int(box['real_time_event']))] is None:
                              remove_candidates[str(int(box['real_time_event']))] = (ycentermean * self.downsamplefactor, xcentermean * self.downsamplefactor)
                        else:
                             remove_candidates_list = remove_candidates[str(int(box['real_time_event']))]
                             remove_candidates_list.append((ycentermean * self.downsamplefactor, xcentermean * self.downsamplefactor))
                             remove_candidates[str(int(box['real_time_event']))] = remove_candidates_list

        #Image back to the same co ordinate system
        self.markers = DownsampleData(self.markers, int(1.0//self.downsamplefactor))
        self.image = DownsampleData(self.image, int(1.0//self.downsamplefactor))

        for i in range(0, self.markers.shape[0]):
                    self.markers[i,:] = self.markers[i,:] > 0
                    self.markers[i,:] = label(self.markers[i,:].astype('uint16'))



        for i in tqdm(range(0, self.markers.shape[0])):
                Clean_Coordinates = []
                sublist = [] 
                waterproperties = measure.regionprops(self.markers[i,:].astype('uint16'))
                sublist = remove_candidates[str(int(i))]
                for prop in waterproperties:
                     current_centroid = prop.centroid
                     
                             
                     if (int(current_centroid[0]), int(current_centroid[1])) not in sublist:
                                
                         Clean_Coordinates.append(current_centroid) 
               
                Clean_Coordinates = sorted(Clean_Coordinates, key=lambda k: [k[1], k[0]])
                Clean_Coordinates.append((0, 0))
                Clean_Coordinates = np.asarray(Clean_Coordinates)
            
                coordinates_int = np.round(Clean_Coordinates).astype(int)
                markers_raw = np.zeros_like(self.markers[i,:])
                markers_raw[tuple(coordinates_int.T)] = 1 + np.arange(len(Clean_Coordinates))
            
                markers = dilation(markers_raw, disk(2))
            
                self.markers[i, :] = label(markers.astype('uint16'))                            
         
        markerdir = self.savedir + '/' + 'Clean_Markers'  
        Path(markerdir).mkdir(exist_ok=True)
        Name = os.path.basename(os.path.splitext(self.imagename)[0])
        print('Writing the clean markers')
        self.marker_tree = MakeTrees(self.markers)
        imwrite(markerdir + '/' + Name + '.tif', self.markers.astype('float32'))    
        



    def second_pass_predict(self):

        print('Detecting event locations')
        eventboxes = []
        refinedeventboxes = []
        classedboxes = {}
        self.n_tiles = (1,1)
        self.iou_threshold = 0.9
        heatsavename = self.savedir+ "/"  + (os.path.splitext(os.path.basename(self.imagename))[0])+ '_Heat'
 
         
        self.image = normalizeFloatZeroOne(self.image.astype('float32'), 1, 99.8)
  
        for inputtime in tqdm(range(0, self.image.shape[0])):
             if inputtime < self.image.shape[0] - self.imaget:   

                if inputtime%(self.image.shape[0]//4)==0 and inputtime > 0 or inputtime >= self.image.shape[0] - self.imaget - 1:



                                      imwrite((heatsavename + '.tif' ), self.heatmap) 
                tree, location = self.marker_tree[str(int(inputtime))]
                for i in range(len(location)):
                    
                    crop_xminus = location[i][1]  - int(self.imagex/2) * self.downsamplefactor - 1
                    crop_xplus = location[i][1]  + int(self.imagex/2) * self.downsamplefactor + 1
                    crop_yminus = location[i][0]  - int(self.imagey/2) * self.downsamplefactor - 1
                    crop_yplus = location[i][0]   + int(self.imagey/2) * self.downsamplefactor + 1
                    region =(slice(inputtime,inputtime + int(self.imaget)),slice(int(crop_yminus), int(crop_yplus)),
                          slice(int(crop_xminus), int(crop_xplus)))
                    
                    crop_image = self.image[region] 
                    if crop_image.shape[0] >= self.imaget and  crop_image.shape[1] >= self.imagey * self.downsamplefactor and crop_image.shape[2] >= self.imagex * self.downsamplefactor:                                                
                                #Now apply the prediction for counting real events
                                
                                crop_image = DownsampleData(crop_image, self.downsamplefactor)
                                ycenter = location[i][0]
                                xcenter = location[i][1]
                                  
                                predictions, allx, ally = self.predict_main(crop_image)
                                #Iterate over tiles
                                for p in range(0,len(predictions)):

                                  sum_time_prediction = predictions[p]
                                  if sum_time_prediction is not None:
                                     #For each tile the prediction vector has shape N H W Categories + Training Vector labels
                                     for i in range(0, sum_time_prediction.shape[0]):
                                          time_prediction =  sum_time_prediction[i]
                                          boxprediction = yoloprediction(0, 0 , time_prediction, self.stride,
                                                           inputtime, self.config,
                                                           self.key_categories, self.key_cord, self.nboxes, 'detection',
                                                           'dynamic')


                                if len(boxprediction) > 0:
                                      for i in range(0, len(boxprediction)):  
                                        boxprediction[i]['xcenter'] = xcenter
                                        boxprediction[i]['ycenter'] = ycenter
                                        boxprediction[i]['xstart'] = xcenter - int(self.imagex/2) * self.downsamplefactor
                                        boxprediction[i]['ystart'] = ycenter - int(self.imagey/2) * self.downsamplefactor  

                                 
                                                 
                                if boxprediction is not None:
                                          eventboxes = eventboxes + boxprediction
                for (event_name,event_label) in self.key_categories.items(): 
                                           
                                        if event_label > 0:
                                             current_event_box = []
                                             for box in eventboxes:
                                        
                                                event_prob = box[event_name]
                                                event_confidence = box['confidence']
                                                if event_prob >= self.event_threshold and event_confidence >= self.event_confidence :
                                                    
                                                             
                                                    current_event_box.append(box)
                                             classedboxes[event_name] = [current_event_box]

                if inputtime > 0:                         
                        self.classedboxes = classedboxes    
                        self.eventboxes =  eventboxes
                        self.nms()
                        self.to_csv()
                        eventboxes = []
                        classedboxes = {}   


    def fast_nms(self):


        best_iou_classedboxes = {}
        self.iou_classedboxes = {}
        for (event_name,event_label) in self.key_categories.items():
            if event_label == 0:

               #best_sorted_event_box = self.classedboxes[event_name][0]
               best_sorted_event_box = dynamic_nms(self.heatmap,self.maskimage, self.classedboxes, event_name,  self.downsamplefactor, self.iou_threshold, self.event_threshold, self.imagex, self.imagey, self.fidelity )

               best_iou_classedboxes[event_name] = [best_sorted_event_box]

        self.iou_classedboxes = best_iou_classedboxes
                
    def nms(self):

        best_iou_classedboxes = {}
        self.iou_classedboxes = {}
        for (event_name,event_label) in self.key_categories.items():
            if event_label > 0:

               #best_sorted_event_box = self.classedboxes[event_name][0]
               if self.remove_markers is not None:
                   best_sorted_event_box = gold_nms(self.heatmap, self.classedboxes, event_name, 1, self.iou_threshold, self.event_threshold, self.imagex, self.imagey, 1 )

               if self.remove_markers == None:
                   best_sorted_event_box = dynamic_nms(self.heatmap,self.maskimage,  self.classedboxes, event_name,  self.downsamplefactor, self.iou_threshold, self.event_threshold, self.imagex, self.imagey, self.fidelity )


               best_iou_classedboxes[event_name] = [best_sorted_event_box]

        self.iou_classedboxes = best_iou_classedboxes



    def to_csv(self):
         if self.remove_markers is not None:
            save_dynamic_csv(self.imagename, self.key_categories, self.iou_classedboxes, self.savedir, 1, self.ndim, z = self.z, maskimage = self.maskimage)        
         if self.markers is None:
            save_dynamic_csv(self.imagename, self.key_categories, self.iou_classedboxes, self.savedir, self.downsamplefactor, self.ndim, z = self.z, maskimage = self.maskimage)          
  

    def saveimage(self, xlocations, ylocations, tlocations, angles, radius, scores):

        # Blue color in BGR
        textcolor = (255, 0, 0)

        # Line thickness of 2 px
        thickness = 2
        for j in range(len(xlocations)):
            startlocation = (int(xlocations[j] - radius[j]), int(ylocations[j] - radius[j]))
            endlocation = (int(xlocations[j] + radius[j]), int(ylocations[j] + radius[j]))
            Z = int(tlocations[j])

            image = self.Colorimage[Z, :, :, 1]
            if scores[j] >= 1.0 - 1.0E-5:
                color = (0, 0, 255)
                image = self.Colorimage[Z, :, :, 2]
            img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            cv2.rectangle(img, startlocation, endlocation, textcolor, thickness)

            cv2.putText(img, str('%.4f' % (scores[j])), startlocation, cv2.FONT_HERSHEY_SIMPLEX, 1, textcolor,
                        thickness, cv2.LINE_AA)
            if scores[j] >= 1.0 - 1.0E-5:
                self.Colorimage[Z, :, :, 2] = img[:, :, 0]
            else:
                self.Colorimage[Z, :, :, 1] = img[:, :, 0]
            if self.yolo_v2:
                x1 = xlocations[j]
                y1 = ylocations[j]

    def showNapari(self, imagedir, savedir, yolo_v2=False):

        Raw_path = os.path.join(imagedir, '*tif')
        X = glob.glob(Raw_path)
        self.savedir = savedir
        Imageids = []
        self.viewer = napari.Viewer()
        napari.run()
        for imagename in X:
            Imageids.append(imagename)

        eventidbox = QComboBox()
        eventidbox.addItem(EventBoxname)
        for (event_name, event_label) in self.key_categories.items():
            eventidbox.addItem(event_name)

        imageidbox = QComboBox()
        imageidbox.addItem(Boxname)
        detectionsavebutton = QPushButton(' Save detection Movie')

        for i in range(0, len(Imageids)):
            imageidbox.addItem(str(Imageids[i]))

        figure = plt.figure(figsize=(4, 4))
        multiplot_widget = FigureCanvas(figure)
        ax = multiplot_widget.figure.subplots(1, 1)
        width = 400
        dock_widget = self.viewer.window.add_dock_widget(
            multiplot_widget, name="EventStats", area='right')
        multiplot_widget.figure.tight_layout()
        self.viewer.window._qt_window.resizeDocks([dock_widget], [width], Qt.Horizontal)
        eventidbox.currentIndexChanged.connect(lambda eventid=eventidbox: EventViewer(
            self.viewer,
            imread(imageidbox.currentText()),
            eventidbox.currentText(),
            self.key_categories,
            os.path.basename(os.path.splitext(imageidbox.currentText())[0]),
            savedir,
            multiplot_widget,
            ax,
            figure,
            yolo_v2,

        )
                                               )

        imageidbox.currentIndexChanged.connect(
            lambda trackid=imageidbox: EventViewer(
                self.viewer,
                imread(imageidbox.currentText()),
                eventidbox.currentText(),
                self.key_categories,
                os.path.basename(os.path.splitext(imageidbox.currentText())[0]),
                savedir,
                multiplot_widget,
                ax,
                figure,
                yolo_v2,

            )
        )

        self.viewer.window.add_dock_widget(eventidbox, name="Event", area='left')
        self.viewer.window.add_dock_widget(imageidbox, name="Image", area='left')

    def overlaptiles(self, sliceregion):

        if self.n_tiles == (1, 1):
            patch = []
            rowout = []
            column = []
            patchx = sliceregion.shape[2] // self.n_tiles[0]
            patchy = sliceregion.shape[1] // self.n_tiles[1]
            patchshape = (patchy, patchx)
            smallpatch, smallrowout, smallcolumn = chunk_list(sliceregion, patchshape, self.stride, [0, 0])
            patch.append(smallpatch)
            rowout.append(smallrowout)
            column.append(smallcolumn)

        else:
            patchx = sliceregion.shape[2] // self.n_tiles[0]
            patchy = sliceregion.shape[1] // self.n_tiles[1]

            if patchx > self.imagex and patchy > self.imagey:
                if self.overlap_percent > 1 or self.overlap_percent < 0:
                    self.overlap_percent = 0.8

                jumpx = int(self.overlap_percent * patchx)
                jumpy = int(self.overlap_percent * patchy)

                patchshape = (patchy, patchx)
                rowstart = 0;
                colstart = 0
                pairs = []
                # row is y, col is x

                while rowstart < sliceregion.shape[1]:
                    colstart = 0
                    while colstart < sliceregion.shape[2]:
                        # Start iterating over the tile with jumps = stride of the fully convolutional network.
                        pairs.append([rowstart, colstart])
                        colstart += jumpx
                    rowstart += jumpy

                    # Include the last patch
                rowstart = sliceregion.shape[1] - patchy
                colstart = 0
                while colstart < sliceregion.shape[2] - patchx:
                    pairs.append([rowstart, colstart])
                    colstart += jumpx
                rowstart = 0
                colstart = sliceregion.shape[2] - patchx
                while rowstart < sliceregion.shape[1] - patchy:
                    pairs.append([rowstart, colstart])
                    rowstart += jumpy

                if sliceregion.shape[1] >= self.imagey and sliceregion.shape[2] >= self.imagex:

                    patch = []
                    rowout = []
                    column = []
                    for pair in pairs:
                        smallpatch, smallrowout, smallcolumn = chunk_list(sliceregion, patchshape, self.stride, pair)
                        if smallpatch.shape[1] >= self.imagey and smallpatch.shape[2] >= self.imagex:
                            patch.append(smallpatch)
                            rowout.append(smallrowout)
                            column.append(smallcolumn)

            else:

                patch = []
                rowout = []
                column = []
                patchx = sliceregion.shape[2] // self.n_tiles[0]
                patchy = sliceregion.shape[1] // self.n_tiles[1]
                patchshape = (patchy, patchx)
                smallpatch, smallrowout, smallcolumn = chunk_list(sliceregion, patchshape, self.stride, [0, 0])
                patch.append(smallpatch)
                rowout.append(smallrowout)
                column.append(smallcolumn)
        self.patch = patch
        self.sy = rowout
        self.sx = column

    def predict_main(self, sliceregion):
        try:
            self.overlaptiles(sliceregion)
            predictions = []
            allx = []
            ally = []
            if len(self.patch) > 0:
                for i in range(0, len(self.patch)):
                    sum_time_prediction = self.make_patches(self.patch[i])
                    predictions.append(sum_time_prediction)
                    allx.append(self.sx[i])
                    ally.append(self.sy[i])


            else:

                sum_time_prediction = self.make_patches(self.patch)
                predictions.append(sum_time_prediction)
                allx.append(self.sx)
                ally.append(self.sy)

        except tf.errors.ResourceExhaustedError:

            print('Out of memory, increasing overlapping tiles for prediction')
            self.list_n_tiles = list(self.n_tiles)
            self.list_n_tiles[0] = self.n_tiles[0] + 1
            self.list_n_tiles[1] = self.n_tiles[1] + 1
            self.n_tiles = tuple(self.list_n_tiles)

            self.predict_main(sliceregion)

        return predictions, allx, ally

    def make_patches(self, sliceregion):

        predict_im = np.expand_dims(sliceregion, 0)

        prediction_vector = self.model.predict(np.expand_dims(predict_im, -1), verbose=0)

        return prediction_vector

    def second_make_patches(self, sliceregion):

        predict_im = np.expand_dims(sliceregion, 0)

        prediction_vector = self.model.predict(np.expand_dims(predict_im, -1), verbose=0)

        return prediction_vector

    def make_batch_patches(self, sliceregion):

        prediction_vector = self.model.predict(np.expand_dims(sliceregion, -1), verbose=0)
        return prediction_vector


def CreateVolume(patch, imaget, timepoint):
    starttime = timepoint
    endtime = timepoint + imaget
    smallimg = patch[starttime:endtime, :]

    return smallimg


def chunk_list(image, patchshape, stride, pair):
    rowstart = pair[0]
    colstart = pair[1]

    endrow = rowstart + patchshape[0]
    endcol = colstart + patchshape[1]

    if endrow > image.shape[1]:
        endrow = image.shape[1]
    if endcol > image.shape[2]:
        endcol = image.shape[2]

    region = (slice(0, image.shape[0]), slice(rowstart, endrow),
              slice(colstart, endcol))

    # The actual pixels in that region.
    patch = image[region]

    # Always normalize patch that goes into the netowrk for getting a prediction score

    return patch, rowstart, colstart


class EventViewer(object):

    def __init__(self, viewer, image, event_name, key_categories, imagename, savedir, canvas, ax, figure, yolo_v2):

        self.viewer = viewer
        self.image = image
        self.event_name = event_name
        self.imagename = imagename
        self.canvas = canvas
        self.key_categories = key_categories
        self.savedir = savedir
        self.ax = ax
        self.yolo_v2 = yolo_v2
        self.figure = figure
        self.plot()

    def plot(self):

        self.ax.cla()

        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0 and self.event_name == event_name:
                csvname = self.savedir + "/" + event_name + "Location" + (
                        os.path.splitext(os.path.basename(self.imagename))[0] + '.csv')
                event_locations, size_locations, angle_locations, line_locations, timelist, eventlist = self.event_counter(
                    csvname)

                for layer in list(self.viewer.layers):
                    if event_name in layer.name or layer.name in event_name or event_name + 'angle' in layer.name or layer.name in event_name + 'angle':
                        self.viewer.layers.remove(layer)
                    if 'Image' in layer.name or layer.name in 'Image':
                        self.viewer.layers.remove(layer)
                self.viewer.add_image(self.image, name='Image')
                self.viewer.add_points(np.asarray(event_locations), size=size_locations, name=event_name,
                                       face_color=[0] * 4, edge_color="red", edge_width=1)
                if self.yolo_v2:
                    self.viewer.add_shapes(np.asarray(line_locations), name=event_name + 'angle', shape_type='line',
                                           face_color=[0] * 4, edge_color="red", edge_width=1)
                self.viewer.theme = 'light'
                self.ax.plot(timelist, eventlist, '-r')
                self.ax.set_title(event_name + "Events")
                self.ax.set_xlabel("Time")
                self.ax.set_ylabel("Counts")
                self.figure.canvas.draw()
                self.figure.canvas.flush_events()
                plt.savefig(self.savedir + event_name + '.png')

    def event_counter(self, csv_file):

        time, y, x, score, size, confidence, angle = np.loadtxt(csv_file, delimiter=',', skiprows=1, unpack=True)

        radius = 10
        eventcounter = 0
        eventlist = []
        timelist = []
        listtime = time.tolist()
        listy = y.tolist()
        listx = x.tolist()
        listsize = size.tolist()
        listangle = angle.tolist()

        event_locations = []
        size_locations = []
        angle_locations = []
        line_locations = []
        for i in range(len(listtime)):
            tcenter = int(listtime[i])
            ycenter = listy[i]
            xcenter = listx[i]
            size = listsize[i]
            angle = listangle[i]
            eventcounter = listtime.count(tcenter)
            timelist.append(tcenter)
            eventlist.append(eventcounter)

            event_locations.append([tcenter, ycenter, xcenter])
            size_locations.append(size)

            xstart = xcenter + radius * math.cos(angle)
            xend = xcenter - radius * math.cos(angle)

            ystart = ycenter + radius * math.sin(angle)
            yend = ycenter - radius * math.sin(angle)
            line_locations.append([[tcenter, ystart, xstart], [tcenter, yend, xend]])
            angle_locations.append(angle)

        return event_locations, size_locations, angle_locations, line_locations, timelist, eventlist

def MidSlices(Image, start_project_mid, end_project_mid, axis = 1):
    
    
    SmallImage = Image.take(indices = range(Image.shape[axis]//2 - start_project_mid, Image.shape[axis]//2 + end_project_mid), axis = axis)
    
    MaxProject = np.amax(SmallImage, axis = axis)
        
    return MaxProject


import sys
sys.path.append("../NEAT")
import csv
import numpy as np
from tifffile import imread, imwrite 
import pandas as pd
import os
import glob
from skimage.measure import regionprops
from skimage import measure
from scipy import spatial 
from tqdm import tqdm
from pathlib import Path
from sklearn.model_selection import train_test_split
from .helpers import  normalizeFloatZeroOne
    
"""
@author: Varun Kapoor
In this program we create training movies and training images for ONEAT. The training data comprises of images and text labels attached to them.
TrainingMovies: This program is for action recognition training data creation. The inputs are the training image, the corresponding integer labelled segmentation image,
csv file containing time, ylocation, xlocation, angle (optional)
Additional parameters to be supplied are the 
1) sizeTminus: action events are centered at the time location, this parameter is the start time of the time volume the network carved out from the image.
2) sizeTplus: this parameter is the end of the time volume to be carved out from the image.
3) total_categories: It is the number of total action categories the network is supposed to predict, Vanilla ONEAT has these labels:
   0: NormalEvent
   1: ApoptosisEvent
   2: DivisionEvent
   3: Macrocheate as static dynamic event
   4: Non MatureP1 cells as static dynamic event
   5: MatureP1 cells as static dynamic event
    
TrainingImages: This program is for cell type recognition training data creation. The inputs are the trainng image, the corresponding integer labelled segmentation image,
Total categories for cell classification part of vanilla ONEAT are:
    0: Normal cells
    1: Central time frame of apoptotic cell
    2: Central time frame of dividing cell
    3: Macrocheates
    4: Non MatureP1 cells
    5: MatureP1 cells
csv file containing time, ylocation, xlocation of that event/cell type
"""    
   
def SegFreeMovieLabelDataSet(image_dir, csv_dir, save_dir, static_name, static_label, csv_name_diff, crop_size, normPatch = False, yolo_v0 = False, yolo_v1 = True, yolo_v2 = False, tshift = 0):
    
    
            raw_path = os.path.join(image_dir, '*tif')
            Csv_path = os.path.join(csv_dir, '*csv')
            files_raw = glob.glob(raw_path)
            files_raw.sort
            filesCsv = glob.glob(Csv_path)
            filesCsv.sort
            Path(save_dir).mkdir(exist_ok=True)
            total_categories = len(static_name)
                
            
            
            for fname in files_raw:
                          
                         name = os.path.basename(os.path.splitext(fname)[0])  
                         
                        
                         for csvfname in filesCsv:
                                 count = 0  
                                 Csvname =  os.path.basename(os.path.splitext(csvfname)[0])
                                 
                                 for i in  range(0, len(static_name)):
                                     event_name = static_name[i]
                                     trainlabel = static_label[i]
                                     classfound = (Csvname == csv_name_diff +  event_name+ name )  
                                     if classfound:
                                                    print(Csvname)
                                                    image = imread(fname)
                                                    if normPatch ==False:
                                                       image = normalizeFloatZeroOne( image.astype('float32'),1,99.8)
                                                    dataset = pd.read_csv(csvfname)
                                                    z = dataset[dataset.keys()[0]][1:]
                                                    y = dataset[dataset.keys()[1]][1:]
                                                    x = dataset[dataset.keys()[2]][1:]
                                                                                
                                                                             
                                                    #Categories + XYHW + Confidence 
                                                    for (key, t) in z.items():
                                                       try: 
                                                          SimpleMovieMaker(t, y[key], x[key], image, crop_size, total_categories, trainlabel, name+ event_name + str(count), save_dir, normPatch,yolo_v0, yolo_v1, yolo_v2, tshift) 
                                                          count = count + 1
                                                        
                                                       except:
                                                        
                                                           pass
                                                        

def SimpleMovieMaker(time, y, x, image, crop_size, total_categories, trainlabel, name, save_dir, normPatch,yolo_v0, yolo_v1, yolo_v2, tshift):
    
       sizex, sizey, size_tminus, size_tplus = crop_size
       
       imagesizex = sizex 
       imagesizey = sizey 
       
       shiftNone = [0,0]
       AllShifts = [shiftNone]


       time = time - tshift
       if time > 0:
               for shift in AllShifts:

                        newname = name + 'shift' + str(shift)
                        Event_data = []
                        
                        if yolo_v0:
                            Label = np.zeros([total_categories + 5])
                        if yolo_v1:    
                            Label = np.zeros([total_categories + 6])
                        if yolo_v2:
                            Label = np.zeros([total_categories + 7])
                        Label[trainlabel] = 1
                        
                        
                        if x + shift[0]> sizex/2 and y + shift[1] > sizey/2 and x + shift[0] + int(imagesizex/2) < image.shape[2] and y + shift[1]+ int(imagesizey/2) < image.shape[1] and time > size_tminus and time + size_tplus + 1 < image.shape[0]:
                                crop_xminus = x  - int(imagesizex/2)
                                crop_xplus = x  + int(imagesizex/2)
                                crop_yminus = y  - int(imagesizey/2)
                                crop_yplus = y   + int(imagesizey/2)
                                # Cut off the region for training movie creation
                                region =(slice(int(time - size_tminus),int(time + size_tplus  + 1)),slice(int(crop_yminus)+ shift[1], int(crop_yplus)+ shift[1]),
                                      slice(int(crop_xminus) + shift[0], int(crop_xplus) + shift[0]))
                                #Define the movie region volume that was cut
                                crop_image = image[region]   
                                if normPatch:
                                    crop_image = normalizeFloatZeroOne( crop_image.astype('float32'),1,99.8)
                                Label[total_categories] = 0.5
                                Label[total_categories + 1] = 0.5
                                Label[total_categories + 2] = 0.5          
                                Label[total_categories + 3] = 0.5
                                Label[total_categories + 4] = 0.5
                                Label[total_categories + 5] = 1
       

                                #Write the image as 32 bit tif file 
                                if(crop_image.shape[0] == size_tplus + size_tminus + 1 and crop_image.shape[1]== imagesizey and crop_image.shape[2]== imagesizex):

                                           imwrite((save_dir + '/' + newname + '.tif'  ) , crop_image.astype('float32'))    
                                           Event_data.append([Label[i] for i in range(0,len(Label))])
                                           if(os.path.exists(save_dir + '/' + (newname) + ".csv")):
                                                        os.remove(save_dir + '/' + (newname) + ".csv")
                                           writer = csv.writer(open(save_dir + '/' + (newname) + ".csv", "a"))
                                           writer.writerows(Event_data)
                                                                    
    
def MovieLabelDataSet(image_dir, seg_image_dir, csv_dir, save_dir, static_name, static_label, csv_name_diff, crop_size, gridx = 1, gridy = 1, offset = 0, yolo_v0 = False, yolo_v1 = True, yolo_v2 = False,  tshift  = 1):
    
    
            raw_path = os.path.join(image_dir, '*tif')
            Seg_path = os.path.join(seg_image_dir, '*tif')
            Csv_path = os.path.join(csv_dir, '*csv')
            files_raw = glob.glob(raw_path)
            files_raw.sort
            filesSeg = glob.glob(Seg_path)
            filesSeg.sort
            filesCsv = glob.glob(Csv_path)
            filesCsv.sort
            Path(save_dir).mkdir(exist_ok=True)
            total_categories = len(static_name)
                
            
            
            for fname in files_raw:
                  
                 name = os.path.basename(os.path.splitext(fname)[0])   
                 for Segfname in filesSeg:
                      
                      Segname = os.path.basename(os.path.splitext(Segfname)[0])
                        
                      if name == Segname:
                          
                          
                         image = imread(fname)
                         image = normalizeFloatZeroOne( image.astype('float32'),1,99.8)
                         segimage = imread(Segfname)
                        
                         for csvfname in filesCsv:
                                 count = 0  
                                 Csvname =  os.path.basename(os.path.splitext(csvfname)[0])
                                 for i in  range(0, len(static_name)):
                                     event_name = static_name[i]
                                     trainlabel = static_label[i]
                                     classfound = (Csvname == csv_name_diff +  event_name + name)   
                                     if classfound:
                                                    print(Csvname)
                                                    dataset = pd.read_csv(csvfname)
                                                    if len(dataset.keys()) >= 3:

                                                        time = dataset[dataset.keys()[0]][1:]
                                                        y = dataset[dataset.keys()[1]][1:]
                                                        x = dataset[dataset.keys()[2]][1:]
                                                        angle = np.full(time.shape, 2)                        
                                                    if len(dataset.keys()) > 3:

                                                        angle = dataset[dataset.keys()[3]][1:]                          
                                                    #Categories + XYHW + Confidence 
                                                    for (key, t) in time.items():
                                                       try: 
                                                          MovieMaker(t, y[key], x[key], angle[key], image, segimage, crop_size, gridx, gridy, offset, total_categories, trainlabel, name + event_name + str(count), save_dir,yolo_v0, yolo_v1, yolo_v2, tshift)
                                                          count = count + 1
                                                        
                                                       except:
                                                        
                                                           pass
                                                        

                                 
                             

               

            
def MovieMaker(time, y, x, angle, image, segimage, crop_size, gridx, gridy, offset, total_categories, trainlabel, name, save_dir, yolo_v0, yolo_v1, yolo_v2, tshift):
    
       sizex, sizey, size_tminus, size_tplus = crop_size
       
       imagesizex = sizex * gridx
       imagesizey = sizey * gridy
       
       shiftNone = [0,0]
       if offset > 0 and trainlabel > 0:
                 shift_lx = [int(offset), 0] 
                 shift_rx = [-offset, 0]
                 shift_lxy = [int(offset), int(offset)]
                 shift_rxy = [-int(offset), int(offset)]
                 shift_dlxy = [int(offset), -int(offset)]
                 shift_drxy = [-int(offset), -int(offset)]
                 shift_uy = [0, int(offset)]
                 shift_dy = [0, -int(offset)]
                 AllShifts = [shiftNone, shift_lx, shift_rx,shift_lxy,shift_rxy,shift_dlxy,shift_drxy,shift_uy,shift_dy]

       else:
           
          AllShifts = [shiftNone]


       time = time - tshift
       if time > 0:
               currentsegimage = segimage[int(time),:].astype('uint16')
               height, width, center, seg_label = getHW(x, y, trainlabel, currentsegimage, imagesizex, imagesizey)
               for shift in AllShifts:

                        newname = name + 'shift' + str(shift)
                        Event_data = []
                        newcenter = (center[0] - shift[1],center[1] - shift[0] )
                        x = center[1]
                        y = center[0]
                        if yolo_v0:
                            Label = np.zeros([total_categories + 5])
                        if yolo_v1:    
                            Label = np.zeros([total_categories + 6])
                        if yolo_v2:
                            Label = np.zeros([total_categories + 7])
                        Label[trainlabel] = 1
                        #T co ordinate
                        Label[total_categories + 2] = (size_tminus) / (size_tminus + size_tplus)
                        if x + shift[0]> sizex/2 and y + shift[1] > sizey/2 and x + shift[0] + int(imagesizex/2) < image.shape[2] and y + shift[1]+ int(imagesizey/2) < image.shape[1] and time > size_tminus and time + size_tplus + 1 < image.shape[0]:
                                        crop_xminus = x  - int(imagesizex/2)
                                        crop_xplus = x  + int(imagesizex/2)
                                        crop_yminus = y  - int(imagesizey/2)
                                        crop_yplus = y   + int(imagesizey/2)
                                        region =(slice(int(time - size_tminus),int(time + size_tplus  + 1)),slice(int(crop_yminus)+ shift[1], int(crop_yplus)+ shift[1]),
                                              slice(int(crop_xminus) + shift[0], int(crop_xplus) + shift[0]))
                                        #Define the movie region volume that was cut
                                        crop_image = image[region]   

                                        seglocationx = (newcenter[1] - crop_xminus)
                                        seglocationy = (newcenter[0] - crop_yminus)

                                        Label[total_categories] =  seglocationx/sizex
                                        Label[total_categories + 1] = seglocationy/sizey
                                        if height >= imagesizey:
                                                        height = 0.5 * imagesizey
                                        if width >= imagesizex:
                                                        width = 0.5 * imagesizex
                                        #Height
                                        Label[total_categories + 3] = height/imagesizey
                                        #Width
                                        Label[total_categories + 4] = width/imagesizex



                                        if yolo_v1:
                                                if seg_label > 0:
                                                  Label[total_categories + 5] = 1 
                                                else:
                                                  Label[total_categories + 5] = 0   

                                        if yolo_v2:

                                             if seg_label > 0:
                                                  Label[total_categories + 5] = 1 
                                             else:
                                                  Label[total_categories + 5] = 0   

                                             Label[total_categories + 6] = angle        

                                        #Write the image as 32 bit tif file 
                                        if(crop_image.shape[0] == size_tplus + size_tminus + 1 and crop_image.shape[1]== imagesizey and crop_image.shape[2]== imagesizex):

                                                   imwrite((save_dir + '/' + newname + '.tif'  ) , crop_image.astype('float32'))    
                                                   Event_data.append([Label[i] for i in range(0,len(Label))])
                                                   if(os.path.exists(save_dir + '/' + (newname) + ".csv")):
                                                                os.remove(save_dir + '/' + (newname) + ".csv")
                                                   writer = csv.writer(open(save_dir + '/' + (newname) + ".csv", "a"))
                                                   writer.writerows(Event_data)
       
   
def Readname(fname):
    
    return os.path.basename(os.path.splitext(fname)[0])


def ImageLabelDataSet(image_dir, seg_image_dir, csv_dir,save_dir, static_name, static_label, csv_name_diff,crop_size, gridx = 1, gridy = 1, offset = 0, yolo_v0 = True, tshift  = 1):
    
    
            raw_path = os.path.join(image_dir, '*tif')
            Seg_path = os.path.join(seg_image_dir, '*tif')
            Csv_path = os.path.join(csv_dir, '*csv')
            files_raw = glob.glob(raw_path)
            files_raw.sort
            filesSeg = glob.glob(Seg_path)
            filesSeg.sort
            filesCsv = glob.glob(Csv_path)
            filesCsv.sort
            Path(save_dir).mkdir(exist_ok=True)
            total_categories = len(static_name)
            
            for csvfname in filesCsv:
              print(csvfname)
              count = 0
              Csvname =  os.path.basename(os.path.splitext(csvfname)[0])
            
              for fname in files_raw:
                  
                 name = os.path.basename(os.path.splitext(fname)[0])   
                 for Segfname in filesSeg:
                      
                      Segname = os.path.basename(os.path.splitext(Segfname)[0])
                        
                      if name == Segname:
                          
                          
                         image = imread(fname)
                         image = normalizeFloatZeroOne( image.astype('float32'),1,99.8)   
                         segimage = imread(Segfname)
                         for i in  range(0, len(static_name)):
                             event_name = static_name[i]
                             trainlabel = static_label[i]
                             if Csvname == csv_name_diff + name + event_name:
                                            dataset = pd.read_csv(csvfname)
                                            time = dataset[dataset.keys()[0]][1:]
                                            y = dataset[dataset.keys()[1]][1:]
                                            x = dataset[dataset.keys()[2]][1:]     
                                            
                                            #Categories + XYHW + Confidence 
                                            for (key, t) in time.items():
                                               ImageMaker(t, y[key], x[key], image, segimage, crop_size, gridx, gridy, offset, total_categories, trainlabel, name + event_name + str(count), save_dir,yolo_v0, tshift)    
                                               count = count + 1
    


    
def SegFreeImageLabelDataSet(image_dir, csv_dir,save_dir, static_name, static_label, csv_name_diff,crop_size, gridx = 1, gridy = 1, offset = 0):
    
    
            raw_path = os.path.join(image_dir, '*tif')
            Csv_path = os.path.join(csv_dir, '*csv')
            files_raw = glob.glob(raw_path)
            files_raw.sort
            filesCsv = glob.glob(Csv_path)
            filesCsv.sort
            Path(save_dir).mkdir(exist_ok=True)
            total_categories = len(static_name)
            
            for csvfname in filesCsv:
              print(csvfname)
              count = 0
              Csvname =  os.path.basename(os.path.splitext(csvfname)[0])
            
              for fname in files_raw:
                  
                         name = os.path.basename(os.path.splitext(fname)[0])   
                         image = imread(fname)
                         image = image[:,:,0]
                         image = normalizeFloatZeroOne( image.astype('float32'),1,99.8)   
                         for i in  range(0, len(static_name)):
                             event_name = static_name[i]
                             trainlabel = static_label[i]
                             if Csvname == csv_name_diff + name + event_name:
                                            dataset = pd.read_csv(csvfname)
                                            #time = dataset[dataset.keys()[0]][1:]
                                            y = dataset[dataset.keys()[1]][1:]
                                            x = dataset[dataset.keys()[0]][1:]     
                                            
                                            #Categories + XYHW + Confidence 
                                            for (key, t) in x.items():
                                               SegFreeImageMaker(1, y[key], x[key], image, crop_size, gridx, gridy, offset, total_categories, trainlabel, name + event_name + str(count), save_dir)    
                                               count = count + 1                 
    
def createNPZ(save_dir, axes, save_name = 'Yolov0oneat', save_name_val = 'Yolov0oneatVal', static = False):
            
            data = []
            label = []   
             
            raw_path = os.path.join(save_dir, '*tif')
            files_raw = glob.glob(raw_path)
            files_raw.sort
            NormalizeImages= [imread(fname) for fname in files_raw]
            
            names = [Readname(fname)  for fname in files_raw]
            #Normalize everything before it goes inside the training
            for i in range(0,len(NormalizeImages)):

                       n = NormalizeImages[i]
                       blankX = n
                       csvfname = save_dir + '/' + names[i] + '.csv'   
                       arr = [] 
                       with open(csvfname) as csvfile:
                             reader = csv.reader(csvfile, delimiter = ',')
                             arr =  list(reader)[0]
                             arr = np.array(arr)
                            
                       blankY = arr
                       
                       blankY = np.expand_dims(blankY, -1)
                       blankX = np.expand_dims(blankX, -1)

                       data.append(blankX)
                       label.append(blankY)



            dataarr = np.asarray(data)
            labelarr = np.asarray(label)
            if static:
                try:
                   dataarr = dataarr[:,0,:,:,:]
                except:
                    
                    pass
            print(dataarr.shape, labelarr.shape)
            traindata, validdata, trainlabel, validlabel = train_test_split(dataarr, labelarr, train_size=0.95,test_size=0.05, shuffle= True)
            save_full_training_data(save_dir, save_name, traindata, trainlabel, axes)
            save_full_training_data(save_dir, save_name_val, validdata, validlabel, axes)


def _raise(e):
    raise e
def  ImageMaker(time, y, x, image, segimage, crop_size, gridX, gridY, offset, total_categories, trainlabel, name, save_dir, yolo_v0, tshift):

               sizeX, sizeY = crop_size

               ImagesizeX = sizeX * gridX
               ImagesizeY = sizeY * gridY

               shiftNone = [0,0]
               if offset > 0 and trainlabel > 0:
                         shiftLX = [int(offset), 0] 
                         shiftRX = [-offset, 0]
                         shiftLXY = [int(offset), int(offset)]
                         shiftRXY = [-int(offset), int(offset)]
                         shiftDLXY = [int(offset), -int(offset)]
                         shiftDRXY = [-int(offset), -int(offset)]
                         shiftUY = [0, int(offset)]
                         shiftDY = [0, -int(offset)]
                         AllShifts = [shiftNone, shiftLX, shiftRX,shiftLXY,shiftRXY,shiftDLXY,shiftDRXY,shiftUY,shiftDY]

               else:

                  AllShifts = [shiftNone]

               time = time - tshift
               if time < segimage.shape[0] - 1 and time > 0:
                 currentsegimage = segimage[int(time),:].astype('uint16')
                
                 height, width, center, SegLabel  = getHW(x, y,trainlabel, currentsegimage, ImagesizeX, ImagesizeY)
                 for shift in AllShifts:
                   
                        newname = name + 'shift' + str(shift)
                        newcenter = (center[0] - shift[1],center[1] - shift[0] )
                        Event_data = []
                        
                        x = center[1]
                        y = center[0]
                        if yolo_v0:
                          Label = np.zeros([total_categories + 4])
                        else:
                          Label = np.zeros([total_categories + 5])  
                        Label[trainlabel] = 1
                        if x + shift[0]> sizeX/2 and y + shift[1] > sizeY/2 and x + shift[0] + int(ImagesizeX/2) < image.shape[2] and y + shift[1]+ int(ImagesizeY/2) < image.shape[1]:
                                    crop_Xminus = x  - int(ImagesizeX/2)
                                    crop_Xplus = x   + int(ImagesizeX/2)
                                    crop_Yminus = y  - int(ImagesizeY/2)
                                    crop_Yplus = y   + int(ImagesizeY/2)
                                    
                                    for tex in range(int(time) -2, int(time) + 2):
                                                    newname = newname + str(tex)
                                                    region =(slice(int(tex - 1),int(tex)),slice(int(crop_Yminus)+ shift[1], int(crop_Yplus)+ shift[1]),
                                                           slice(int(crop_Xminus) + shift[0], int(crop_Xplus) + shift[0]))

                                                    crop_image = image[region]      


                                                    seglocationX = (newcenter[1] - crop_Xminus)
                                                    seglocationY = (newcenter[0] - crop_Yminus)

                                                    Label[total_categories] =  seglocationX/sizeX
                                                    Label[total_categories + 1] = seglocationY/sizeY

                                                    if height >= ImagesizeY:
                                                        height = 0.5 * ImagesizeY
                                                    if width >= ImagesizeX:
                                                        width = 0.5 * ImagesizeX

                                                    Label[total_categories + 2] = height/ImagesizeY
                                                    Label[total_categories + 3] = width/ImagesizeX




                                                    if yolo_v0==False:
                                                            if SegLabel > 0:
                                                              Label[total_categories + 4] = 1 
                                                            else:
                                                              Label[total_categories + 4] = 0  

                                                    if(crop_image.shape[1]== ImagesizeY and crop_image.shape[2]== ImagesizeX):
                                                             imwrite((save_dir + '/' + newname + '.tif'  ) , crop_image.astype('float32'))  
                                                             Event_data.append([Label[i] for i in range(0,len(Label))])
                                                             if(os.path.exists(save_dir + '/' + (newname) + ".csv")):
                                                                os.remove(save_dir + '/' + (newname) + ".csv")
                                                             writer = csv.writer(open(save_dir + '/' + (newname) + ".csv", "a"))
                                                             writer.writerows(Event_data)

def  SegFreeImageMaker(time, y, x, image, crop_size, gridX, gridY, offset, total_categories, trainlabel, name, save_dir):

               sizeX, sizeY = crop_size

               ImagesizeX = sizeX * gridX
               ImagesizeY = sizeY * gridY

               shiftNone = [0,0]
               if offset > 0 and trainlabel > 0:
                         shiftLX = [int(offset), 0] 
                         shiftRX = [-offset, 0]
                         shiftLXY = [int(offset), int(offset)]
                         shiftRXY = [-int(offset), int(offset)]
                         shiftDLXY = [int(offset), -int(offset)]
                         shiftDRXY = [-int(offset), -int(offset)]
                         shiftUY = [0, int(offset)]
                         shiftDY = [0, -int(offset)]
                         AllShifts = [shiftNone, shiftLX, shiftRX,shiftLXY,shiftRXY,shiftDLXY,shiftDRXY,shiftUY,shiftDY]

               else:

                  AllShifts = [shiftNone]

               center = (y,x)
               
               for shift in AllShifts:
                   
                        newname = name + 'shift' + str(shift)
                        newcenter = (center[0] - shift[1],center[1] - shift[0] )
                        Event_data = []
                        
                        x = center[1]
                        y = center[0]
                        Label = np.zeros([total_categories + 2])
                         
                        Label[trainlabel] = 1
                        if x + shift[0]> sizeX/2 and y + shift[1] > sizeY/2 and x + shift[0] + int(ImagesizeX/2) < image.shape[1] and y + shift[1]+ int(ImagesizeY/2) < image.shape[0]:
                                    crop_Xminus = x  - int(ImagesizeX/2)
                                    crop_Xplus = x   + int(ImagesizeX/2)
                                    crop_Yminus = y  - int(ImagesizeY/2)
                                    crop_Yplus = y   + int(ImagesizeY/2)
                                 
                                    
                                    region =(slice(int(crop_Yminus)+ shift[1], int(crop_Yplus)+ shift[1]),
                                           slice(int(crop_Xminus) + shift[0], int(crop_Xplus) + shift[0]))
                                    crop_image = image[region]      
                                    seglocationX = (newcenter[1] - crop_Xminus)
                                    seglocationY = (newcenter[0] - crop_Yminus)
                                      
                                    Label[total_categories] =  seglocationX/sizeX
                                    Label[total_categories + 1] = seglocationY/sizeY
                                    

                                    
                                    if(crop_image.shape[0]== ImagesizeY and crop_image.shape[1]== ImagesizeX):
                                             imwrite((save_dir + '/' + newname + '.tif'  ) , crop_image.astype('float32'))  
                                             Event_data.append([Label[i] for i in range(0,len(Label))])
                                             if(os.path.exists(save_dir + '/' + (newname) + ".csv")):
                                                os.remove(save_dir + '/' + (newname) + ".csv")
                                             writer = csv.writer(open(save_dir + '/' + (newname) + ".csv", "a"))
                                             writer.writerows(Event_data)
       
def getHW(defaultX, defaultY, trainlabel, currentsegimage, imagesizex, imagesizey):
    
    properties = measure.regionprops(currentsegimage, currentsegimage)
    TwoDLocation = (defaultY,defaultX)
    TwoDCoordinates = [(prop.centroid[0], prop.centroid[1]) for prop in properties]
    SegLabel = currentsegimage[int(TwoDLocation[0]), int(TwoDLocation[1])]
    for prop in properties:
                                               
                  if SegLabel > 0 and prop.label == SegLabel:
                                    minr, minc, maxr, maxc = prop.bbox
                                    center = (defaultY, defaultX)
                                    height =  abs(maxc - minc)
                                    width =  abs(maxr - minr)
                                
                  if SegLabel == 0 :
                    
                             center = (defaultY, defaultX)
                             height = 0.5 * imagesizex
                             width = 0.5 * imagesizey
                               
                    
                                
    return height, width, center, SegLabel     

def save_full_training_data(directory, filename, data, label, axes):
    """Save training data in ``.npz`` format."""
  

    len(axes) == data.ndim or _raise(ValueError())
    np.savez(directory + filename, data = data, label = label, axes = axes)
    
def InterchangeTXY(TXYCSV, save_dir):

    
     dataset = pd.read_csv(TXYCSV)
     time = dataset[dataset.keys()[0]][1:]
     x = dataset[dataset.keys()[1]][1:]
     y = dataset[dataset.keys()[2]][1:]
 
     Event_data = []
     
     Name = os.path.basename(os.path.splitext(TXYCSV)[0])

     for (key, t) in time.items():
         
         Event_data.append([t, y[key], x[key]])
         
     writer = csv.writer(open(save_dir + '/' + (Name) + ".csv", "a"))
     writer.writerows(Event_data)    
         
    
def  AngleAppender(AngleCSV, ONTCSV, save_dir, ColumnA = 'Y'):

     dataset = pd.read_csv(AngleCSV)
     time = dataset[dataset.keys()[0]][1:]
     if ColumnA == 'Y':
       y = dataset[dataset.keys()[1]][1:]
       x = dataset[dataset.keys()[2]][1:]
       angle = dataset[dataset.keys()[3]][1:]
     else:
       x = dataset[dataset.keys()[1]][1:]
       y = dataset[dataset.keys()[2]][1:]
       angle = dataset[dataset.keys()[3]][1:]  
       
     clickeddataset = pd.read_csv(ONTCSV)  
     
     clickedtime = clickeddataset[clickeddataset.keys()[0]][1:]
     clickedy = clickeddataset[clickeddataset.keys()[1]][1:]
     clickedx = clickeddataset[clickeddataset.keys()[2]][1:]
                              
     Event_data = []
     
     Name = os.path.basename(os.path.splitext(ONTCSV)[0])
     

     
     for (clickedkey, clickedt) in clickedtime.items():
                
                       
                for (key, t) in time.items(): 
                         
                          if t == clickedt and y[key] == clickedy[clickedkey] and x[key] == clickedx[clickedkey]:
                              
                                Event_data.append([t,y[key],x[key],angle[key]])
                              
                                     
                              
                               
     writer = csv.writer(open(save_dir + '/' + (Name) + ".csv", "a"))
     writer.writerows(Event_data)
         



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 28 13:49:35 2021

@author: vkapoor
"""

from .neat_goldstandard import NEATDynamic

class NEATSynamic(NEATDynamic):
    

    def __init__(self, config, model_dir, model_name, catconfig = None, cordconfig = None):

        super().__init__(config = config, model_dir = model_dir, model_name = model_name, catconfig = catconfig, cordconfig = cordconfig)
        
    def predict_synamic(self,imagename, savedir, n_tiles = (1,1), overlap_percent = 0.8, event_threshold = 0.5, event_confidence = 0.5, iou_threshold = 0.1, fidelity = 5,
     downsamplefactor = 1, maskfilter = 10,start_project_mid = 4, end_project_mid = 4, maskmodel = None, maskdir = None):

        self.predict(imagename,savedir,n_tiles = n_tiles, overlap_percent = overlap_percent, event_threshold = event_threshold, event_confidence = event_confidence, iou_threshold = iou_threshold, 
        fidelity = fidelity, start_project_mid = start_project_mid, end_project_mid = end_project_mid, downsamplefactor = downsamplefactor, maskfilter = maskfilter,  markers = None, marker_tree = None,
         remove_markers = None, maskmodel = maskmodel, maskdir = maskdir )


import random
import argparse
import numpy as np
import os
import json
import glob
import csv

def IOU(ann, centroids):
    w, h = ann
    similarities = []

    for centroid in centroids:
        c_w, c_h = centroid

        if c_w >= w and c_h >= h:
            similarity = w*h/(c_w*c_h)
        elif c_w >= w and c_h <= h:
            similarity = w*c_h/(w*h + (c_w-w)*c_h)
        elif c_w <= w and c_h >= h:
            similarity = c_w*h/(w*h + c_w*(c_h-h))
        else: #means both w,h are bigger than c_w and c_h respectively
            similarity = (c_w*c_h)/(w*h)
        similarities.append(similarity) # will become (k,) shape

    return np.array(similarities)

def avg_IOU(anns, centroids):
    n,d = anns.shape
    sum = 0.

    for i in range(anns.shape[0]):
        sum+= max(IOU(anns[i], centroids))

    return sum/n

def print_anchors(centroids):
    anchors = centroids.copy()

    widths = anchors[:, 0]
    sorted_indices = np.argsort(widths)

    r = "anchors: ["
    for i in sorted_indices[:-1]:
        r += '%0.2f,%0.2f, ' % (anchors[i,0], anchors[i,1])

    #there should not be comma after last anchor, that's why
    r += '%0.2f,%0.2f' % (anchors[sorted_indices[-1:],0], anchors[sorted_indices[-1:],1])
    r += "]"

    print(r)

def run_kmeans(ann_dims, anchor_num):
    ann_num = ann_dims.shape[0]
    iterations = 0
    prev_assignments = np.ones(ann_num)*(-1)
    iteration = 0
    old_distances = np.zeros((ann_num, anchor_num))

    indices = [random.randrange(ann_dims.shape[0]) for i in range(anchor_num)]
    centroids = ann_dims[indices]
    anchor_dim = ann_dims.shape[1]

    while True:
        distances = []
        iteration += 1
        for i in range(ann_num):
            d = 1 - IOU(ann_dims[i], centroids)
            distances.append(d)
        distances = np.array(distances) # distances.shape = (ann_num, anchor_num)

        print("iteration {}: dists = {}".format(iteration, np.sum(np.abs(old_distances-distances))))

        #assign samples to centroids
        assignments = np.argmin(distances,axis=1)

        if (assignments == prev_assignments).all() :
            return centroids

        #calculate new centroids
        centroid_sums=np.zeros((anchor_num, anchor_dim), np.float)
        for i in range(ann_num):
            centroid_sums[assignments[i]]+=ann_dims[i]
        for j in range(anchor_num):
            centroids[j] = centroid_sums[j]/(np.sum(assignments==j) + 1e-6)

        prev_assignments = assignments.copy()
        old_distances = distances.copy()

def main():
    CSVDir = '/data/u934/service_imagerie/v_kapoor/FinalONEATTraining/Bin2OYoloneatV1/'
    Csv_path = os.path.join(CSVDir, '*csv')
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    num_anchors = 5


    grid_w = 1
    grid_h = 1
    categories = 4
    # run k_mean to find the anchors
    annotation_dims = []
    for csvfname in filesCsv:


             with open(csvfname) as csvfile:
                     reader = csv.reader(csvfile, delimiter = ',')
                     for train_vec in reader:

                             arr =  [float(s) for s in train_vec[0:]]
             train_vec = arr
            
             
             relative_w = float(train_vec[6])
             relative_h = float(train_vec[7])
             
             annotation_dims.append(tuple(map(float, (relative_w,relative_h))))

    annotation_dims = np.array(annotation_dims)
    centroids = run_kmeans(annotation_dims, num_anchors)

    # write anchors to file
    print('\naverage IOU for', num_anchors, 'anchors:', '%0.2f' % avg_IOU(annotation_dims, centroids))
    print_anchors(centroids)

if __name__ == '__main__':
    
    main()
